/*
**
**  MODULE DESCRIPTION:
**
**      This program is an example application to make slides
**	for presentations.
**
**	Copyright (c) 1991-1993 Sebern Engineering Inc.
**	All rights reserved.
**
**  AUTHOR:
**
**      Mark Sebern
**	Sebern Engineering Inc.
**	P. O. Box 268
**	Cedarburg, WI 53012
**	414/375-2200
**
**  CREATION DATE:  18-Feb-1991
**
*/


/*
**
**  INCLUDE FILES
**
*/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <limits.h>
#ifdef VMS
#pragma nostandard
#endif
#include <Mrm/MrmAppl.h>
#include <X11/X.h>
#include <X11/cursorfont.h>
#include <X11/keysym.h>
#include <Xm/FileSB.h>
#include <Xm/List.h>
#include <Xm/MessageB.h>
#include <Xm/PushB.h>
#include <Xm/RowColumn.h>
#include <Xm/Scale.h>
#include <Xm/SelectioB.h>
#include <Xm/Text.h>
#include <Xm/TextF.h>
#include <Xm/ToggleB.h>
#ifdef VMS
#pragma standard
#endif

/*
**
**  MACROS
**
*/

#ifndef MIN
#define MIN(x,y) ((x) < (y) ? (x) : (y))
#endif /* MIN */

#ifndef MAX
#define MAX(x,y) ((x) > (y) ? (x) : (y))
#endif /* MAX */

/*  Cursor shape index to specify default cursor. */
#define DEFAULT_CURSOR	    (UINT_MAX)

/*  Define this macro to use the UIL-declared application icon. */
/**
#define APP_ICON_UIL	    0
**/

/*  Application-specific resource instance and class names: */
#define MslideNbufferPixmap	"bufferPixmap"
#define MslideCBufferPixmap	"BufferPixmap"
#define MslideNclearBackground	"clearBackground"
#define MslideCClearBackground	"ClearBackground"
#define MslideNcolorNumColumns	"colorNumColumns"
#define MslideCColorNumColumns	"ColorNumColumns"
#define MslideNdebugFlag	"debugFlag"
#define MslideCDebugFlag	"DebugFlag"
#define MslideNpaletteSize	"paletteSize"
#define MslideCPaletteSize	"PaletteSize"

#define PALETTE_SIZE_DEFAULT	6
#define PALETTE_SIZE_MIN	2

#define COLOR_NUM_COLS_DEFAULT	1

/*
**
**  DATA TYPE DEFINITIONS
**
*/

/*  The data structure used to associate widget names and widget
    ID values, so the widget ID can be looked up by name. An
    optional index value can be used to distinguish widgets that
    are assigned the same name. */

typedef struct widget_entry {
    String		wid_name;
    int			wid_index;
    Widget		wid_id;
    struct widget_entry	*next;
} widget_entry_type;

/*  Maximum number of "body lines" in a slide. */
#define SLIDE_BODY_MAX	5

/*  The "body line" data structure. It specifies a line of
    body text and an optional "bullet" symbol. */

typedef struct {
    Boolean		body_bullet;
    String		body_text;
} slide_body_type;

/*  The slide data structure. This structure contains the
    information to specify one slide in a slide show. */

typedef struct {
    String		master_name;
    String		header_text;
    Boolean		header_bar;
    String		title_text;
    slide_body_type	body_lines [SLIDE_BODY_MAX];
} slide_info_type;

/*  The master element data structure. This structure contains
    the information necessary to specify one graphic element in
    a slide master. */

typedef struct {
    unsigned int	element_type;
    int			element_x;
    int			element_y;
    unsigned int	element_width;
    unsigned int	element_height;
    String		element_font_name;
    String		element_pixmap_name;
    Cardinal		element_foreground;
    Cardinal		element_background;
    int			element_data;
    int			element_index;
    Pixmap		element_pixmap;
    GC			element_gc;
    Widget		element_gc_widget;
    Font		element_font_id;
    XtGCMask		element_gc_mask;
} master_element_type;

/*  Values for "element_type" member of the master element
    data structure. */

#define EL_TYPE_END	    0
#define EL_TYPE_MIN	    1
#define EL_TYPE_RECT		1   /* Rectangle */
#define EL_TYPE_TEXT		2   /* Text string */
#define EL_TYPE_TEXT_STAR	3   /* Text string (highlighted) */
#define EL_TYPE_BULLET		4   /* Bullet */
#define EL_TYPE_PIXMAP		5   /* Pixmap */
#define EL_TYPE_MAX	    5

static String element_type_names [] =
{
    "End",
    "Rectangle",
    "Text",
    "Text*",
    "Bullet",
    "Pixmap",
    (String) NULL
};

/*  Values for "element_data" member of the master element
    data structure. One of these constants specifies the
    source of data for a given element, if one is needed.
    For example, a text string element has an associated
    text data value (header, title, body), while a rectangle,
    bullet, or pixmap may be conditioned by a Boolean value
    (header bar or body bullet). Some data sources (body text,
    body bullet) are indexed by the "element_index" member of
    the element data structure. */

#define EL_DATA_NONE	    0
#define EL_DATA_HEADER_TEXT	1	/* Header text string */
#define EL_DATA_HEADER_BAR	2	/* Header bar Boolean */
#define EL_DATA_TITLE_TEXT	3	/* Title text string */
#define EL_DATA_BODY_TEXT	4	/* Body text string */
#define EL_DATA_BODY_BULLET	5	/* Body bullet Boolean */
#define EL_DATA_MAX	    5

/*  The following definitions set up a general mechanism for
    handling aggregates of data structures. This mechanism is
    used to keep track of slides in a slide show, graphic
    elements in a slide master, and slide masters in a slide
    master list. The "member functions" (fcn_create and fcn_copy)
    are called to create a new record or copy a record, of the
    particular type defined for the aggregate. */

typedef void (*reclist_copy_fcn) (void *drec, void *srec);
typedef void *(*reclist_create_fcn) (void);

typedef struct {
    unsigned int	    count;
    unsigned int	    select_index;
    unsigned int	    max_count;
    Boolean		    modified_flag;
    void		    **recs;
    reclist_create_fcn	    fcn_create;
    reclist_copy_fcn	    fcn_copy;
} reclist_type;

#define RECLIST_ALLOC_DELTA 	20

/*  Slide master data structure. A slide master consists of
    a slide master name and an ordered set of graphic element
    definitions. */

typedef struct {
    String		    name;
    reclist_type	    *elem_list;
} slide_master_type;


/*  This data structure describes one color in the
    color palette. */
typedef struct {
    String		    color_name;
    Pixel		    pixel;
    Boolean		    allocated_pixel;
    Pixmap		    sample_pixmap;
} palette_entry_type;

/*  This data structure describes the entire
    color palette. */
typedef struct {
    Cardinal		    size;
    Boolean		    modified_flag;
    Screen		    *screen;
    palette_entry_type	    *entries;
} palette_type;

/*  A data structure to maintain processing state for the
    "print show" work procedure. */

struct print_show_info
{
    String		filename;
    reclist_type	*show_ptr;
    reclist_type	*master_list_ptr;
    FILE		*filep;
    unsigned int	slide_index;
};

/*  A data structure to maintain processing state for the
    "read show" work procedure. */

struct read_show_info
{
    String		filename;
    reclist_type	*show_ptr;
    palette_type	*color_palette_ptr;
    reclist_type	*master_list_ptr;
    FILE		*filep;
    slide_info_type	*new_slide;
    slide_master_type	*new_master;
    master_element_type	*new_element;
};

/*  A data structure to maintain processing state for the
    "save show" work procedure. */

struct write_show_info
{
    String		filename;
    reclist_type	*show_ptr;
    palette_type	*color_palette_ptr;
    reclist_type	*master_list_ptr;
    FILE		*filep;
    unsigned int	color_index;
    unsigned int	master_index;
    unsigned int	slide_index;
};

/*  This is the definition of the application resource data structure.
    A variable of this type receives the values for application
    resources, derived from the toolkit resource database. */
typedef struct {
    /*	Flag to indicate whether we should draw the slide image to
	a pixmap and then refresh the drawing window from the
	pixmap. Otherwise, we just draw directly to the screen. */
    Boolean	use_buffer_pixmap;

    /*	Flag to indicate whether we should clear the entire slide
	background before drawing a new slide. */
    Boolean	clear_bkgnd;

    /*	The debug flag: if set, the application displays messages
	to trace the progress of its execution. */
    Boolean	debug;

    /*	The number of entries in the color palette. We actually want
	a "Cardinal" value here instead of an "int", but there is
	no predefined type converter for "String" to "Cardinal". */
    int		palette_size;

    /*	The number of columns used to display the color samples
	in the foreground/background color option menus. If set
	greater than one, it permits the display of more color
	palette entry samples without falling off the screen. */
    int		color_num_cols;
} application_resource_type;

/*
**
**  DATA DECLARATIONS
**
*/

static MrmHierarchy	s_MrmHierarchy;
static String		uid_file_names [] =
{
    "mslide.uid"
};
static MrmCount		uid_file_num = XtNumber (uid_file_names);

static XtAppContext	app_context;
static Widget		wid_top_level;
static Widget		wid_main_window = (Widget) NULL;
static Display		*main_display;

static widget_entry_type *wid_list_head =
		(widget_entry_type *) NULL;

static Boolean		debug_flag;

static reclist_type	*main_show;

static slide_master_type    *edit_master;
static slide_master_type    *null_master;

static reclist_type	*main_master_list;

static String		current_file_name = (String) NULL;
static String		app_name = (String) "Mslide";
static String		help_shell_title = (String) "Mslide Help";

static String		ks_slide_list_descr_sel;
static String		ks_slide_list_descr_nosel;
static Pixel		slide_descr_pixel_sel;
static Pixel		slide_descr_pixel_nosel;

static slide_info_type	*disp_slide;
static unsigned int	disp_indx;
static GC		draw_clear_gc = (GC) NULL;
static Pixmap		buffer_pixmap;

static String		check_save_tag = (String) NULL;
static String		overwrite_file_name = (String) NULL;

/*  Color palette pointer. */
static palette_type	*main_palette = (palette_type *) NULL;

static char		default_header_font [] =
	"-*-Helvetica-Medium-R-Normal-*-*-180-100-100-*-*-*-*";
static char		default_title_font [] =
	"-*-Helvetica-Bold-R-Normal-*-*-240-100-100-*-*-*-*";
static char		default_body_font [] =
	"-*-Helvetica-Bold-R-Normal-*-*-240-100-100-*-*-*-*";

#define DEFAULT_BACKGROUND_COLOR    1
#define DEFAULT_HEADER_COLOR	    2
#define DEFAULT_TITLE_COLOR	    3
#define DEFAULT_BODY_COLOR	    4

static String		default_color_names [] =
{
    (String) "#C51D30A3451E",
    (String) "#FFFF999975C2",
    (String) "#5C28A8F568F5",
    (String) "#FFFFF0A2DC28",
    (String) NULL
};

static master_element_type  default_title_master [] = {
{ EL_TYPE_RECT, 150, 150, 600, 400, (String) NULL, (String) NULL,
  DEFAULT_BACKGROUND_COLOR, 0, EL_DATA_NONE, 0,
  XmUNSPECIFIED_PIXMAP, (GC) NULL, (Widget) NULL, (Font) NULL, 0 },
{ EL_TYPE_TEXT, 180, 230, 0, 0, (String) default_title_font,
  (String) NULL, DEFAULT_TITLE_COLOR, 0, EL_DATA_TITLE_TEXT, 0,
  XmUNSPECIFIED_PIXMAP, (GC) NULL, (Widget) NULL, (Font) NULL, 0 },
{ EL_TYPE_TEXT, 220, 290, 0, 0, (String) default_body_font,
  (String) NULL, DEFAULT_BODY_COLOR, 0, EL_DATA_BODY_TEXT , 1,
  XmUNSPECIFIED_PIXMAP, (GC) NULL, (Widget) NULL, (Font) NULL, 0 },
{ EL_TYPE_TEXT, 220, 345, 0, 0, (String) default_body_font,
  (String) NULL, DEFAULT_BODY_COLOR, 0, EL_DATA_BODY_TEXT , 2,
  XmUNSPECIFIED_PIXMAP, (GC) NULL, (Widget) NULL, (Font) NULL, 0 },
{ EL_TYPE_TEXT, 220, 400, 0, 0, (String) default_body_font,
  (String) NULL, DEFAULT_BODY_COLOR, 0, EL_DATA_BODY_TEXT , 3,
  XmUNSPECIFIED_PIXMAP, (GC) NULL, (Widget) NULL, (Font) NULL, 0 },
{ EL_TYPE_TEXT, 220, 455, 0, 0, (String) default_body_font,
  (String) NULL, DEFAULT_BODY_COLOR, 0, EL_DATA_BODY_TEXT , 4,
  XmUNSPECIFIED_PIXMAP, (GC) NULL, (Widget) NULL, (Font) NULL, 0 },
{ EL_TYPE_TEXT, 220, 510, 0, 0, (String) default_body_font,
  (String) NULL, DEFAULT_BODY_COLOR, 0, EL_DATA_BODY_TEXT , 5,
  XmUNSPECIFIED_PIXMAP, (GC) NULL, (Widget) NULL, (Font) NULL, 0 },
{ EL_TYPE_END, 0, 0, 0, 0, (String) NULL, (String) NULL, 0, 0,
  EL_DATA_NONE, 0, XmUNSPECIFIED_PIXMAP, (GC) NULL, (Widget) NULL,
  (Font) NULL, 0 },
};

static master_element_type default_body_master [] = {
{ EL_TYPE_RECT, 150, 150, 600, 400, (String) NULL, (String) NULL,
  DEFAULT_BACKGROUND_COLOR, 0, EL_DATA_NONE, 0,
  XmUNSPECIFIED_PIXMAP, (GC) NULL, (Widget) NULL, (Font) NULL, 0 },
{ EL_TYPE_TEXT, 180, 185, 0, 0, (String) default_header_font,
  (String) NULL, DEFAULT_HEADER_COLOR, 0, EL_DATA_HEADER_TEXT, 0,
  XmUNSPECIFIED_PIXMAP, (GC) NULL, (Widget) NULL, (Font) NULL, 0 },
{ EL_TYPE_RECT, 170, 195, 560, 2, (String) NULL, (String) NULL,
  DEFAULT_HEADER_COLOR, 0, EL_DATA_HEADER_BAR, 0,
  XmUNSPECIFIED_PIXMAP, (GC) NULL, (Widget) NULL, (Font) NULL, 0 },
{ EL_TYPE_TEXT, 180, 230, 0, 0, (String) default_title_font,
  (String) NULL, DEFAULT_TITLE_COLOR, 0, EL_DATA_TITLE_TEXT, 0,
  XmUNSPECIFIED_PIXMAP, (GC) NULL, (Widget) NULL, (Font) NULL, 0 },
{ EL_TYPE_TEXT, 220, 290, 0, 0, (String) default_body_font,
  (String) NULL, DEFAULT_BODY_COLOR, 0, EL_DATA_BODY_TEXT, 1,
  XmUNSPECIFIED_PIXMAP, (GC) NULL, (Widget) NULL, (Font) NULL, 0 },
{ EL_TYPE_TEXT, 220, 345, 0, 0, (String) default_body_font,
  (String) NULL, DEFAULT_BODY_COLOR, 0, EL_DATA_BODY_TEXT, 2,
  XmUNSPECIFIED_PIXMAP, (GC) NULL, (Widget) NULL, (Font) NULL, 0 },
{ EL_TYPE_TEXT, 220, 400, 0, 0, (String) default_body_font,
  (String) NULL, DEFAULT_BODY_COLOR, 0, EL_DATA_BODY_TEXT, 3,
  XmUNSPECIFIED_PIXMAP, (GC) NULL, (Widget) NULL, (Font) NULL, 0 },
{ EL_TYPE_TEXT, 220, 455, 0, 0, (String) default_body_font,
  (String) NULL, DEFAULT_BODY_COLOR, 0, EL_DATA_BODY_TEXT, 4,
  XmUNSPECIFIED_PIXMAP, (GC) NULL, (Widget) NULL, (Font) NULL, 0 },
{ EL_TYPE_TEXT, 220, 510, 0, 0, (String) default_body_font,
  (String) NULL, DEFAULT_BODY_COLOR, 0, EL_DATA_BODY_TEXT, 5,
  XmUNSPECIFIED_PIXMAP, (GC) NULL, (Widget) NULL, (Font) NULL, 0 },
{ EL_TYPE_BULLET, 190, 270, 15, 15, (String) NULL, (String) NULL,
  DEFAULT_BODY_COLOR, 0, EL_DATA_BODY_BULLET, 1,
  XmUNSPECIFIED_PIXMAP, (GC) NULL, (Widget) NULL, (Font) NULL, 0 },
{ EL_TYPE_BULLET, 190, 325, 15, 15, (String) NULL, (String) NULL,
  DEFAULT_BODY_COLOR, 0, EL_DATA_BODY_BULLET, 2,
  XmUNSPECIFIED_PIXMAP, (GC) NULL, (Widget) NULL, (Font) NULL, 0 },
{ EL_TYPE_BULLET, 190, 380, 15, 15, (String) NULL, (String) NULL,
  DEFAULT_BODY_COLOR, 0, EL_DATA_BODY_BULLET, 3,
  XmUNSPECIFIED_PIXMAP, (GC) NULL, (Widget) NULL, (Font) NULL, 0 },
{ EL_TYPE_BULLET, 190, 435, 15, 15, (String) NULL,
  (String) NULL, DEFAULT_BODY_COLOR, 0, EL_DATA_BODY_BULLET, 4,
  XmUNSPECIFIED_PIXMAP, (GC) NULL, (Widget) NULL, (Font) NULL, 0 },
{ EL_TYPE_BULLET, 190, 490, 15, 15, (String) NULL, (String) NULL,
  DEFAULT_BODY_COLOR, 0, EL_DATA_BODY_BULLET, 5,
  XmUNSPECIFIED_PIXMAP, (GC) NULL, (Widget) NULL, (Font) NULL, 0 },
{ EL_TYPE_END, 0, 0, 0, 0, (String) NULL, (String) NULL, 0, 0,
  EL_DATA_NONE, 0, XmUNSPECIFIED_PIXMAP, (GC) NULL, (Widget) NULL,
  (Font) NULL, 0 },
};

#ifndef APP_ICON_UIL
/*  If we are not using the UIL-declared application icon,
    we specify it in binary here, one bit per pixel. */

#define APP_ICON_WIDTH	32
#define APP_ICON_HEIGHT	32

static char		    app_icon_data [] =
{
    0xfc,0xff,0xff,0x3f,    0xfe,0xff,0xff,0x7f,
    0x07,0x00,0x00,0xe0,    0x03,0x00,0x00,0xc0,
    0x03,0x00,0x00,0xc0,    0x03,0x00,0x00,0xc0,
    0x03,0x00,0x00,0xc0,    0x03,0x00,0x00,0xc0,
    0xe3,0xff,0xff,0xc7,    0xa3,0xaa,0xaa,0xc6,
    0x63,0x55,0x55,0xc5,    0xa3,0xaa,0xaa,0xc6,
    0x63,0x55,0x55,0xc5,    0xa3,0xaa,0xaa,0xc6,
    0x63,0x55,0x55,0xc5,    0xa3,0xaa,0xaa,0xc6,
    0x63,0x55,0x55,0xc5,    0xa3,0xaa,0xaa,0xc6,
    0x63,0x55,0x55,0xc5,    0xa3,0xaa,0xaa,0xc6,
    0x63,0x55,0x55,0xc5,    0xa3,0xaa,0xaa,0xc6,
    0x63,0x55,0x55,0xc5,    0xe3,0xff,0xff,0xc7,
    0x03,0x00,0x00,0xc0,    0x03,0x00,0x00,0xc0,
    0x03,0x00,0x00,0xc0,    0x03,0x00,0x00,0xc0,
    0x03,0x00,0x00,0xc0,    0x07,0x00,0x00,0xe0,
    0xfe,0xff,0xff,0x7f,    0xfc,0xff,0xff,0x3f,
};
#endif

/*  Logo image definition: This pattern specifies a multi-color image
    that can be converted to a pixmap and displayed as part of a
    slide. By rearranging the data, a more useful pattern (like a
    company logo) could be created. */
#define LOGO_IMAGE_NAME	    ((String) "logo")
#define LOGO_IMAGE_FORMAT   (ZPixmap)
#define LOGO_IMAGE_WIDTH    32
#define LOGO_IMAGE_HEIGHT   32
static unsigned char	logo_image_data [] =
{
    2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,
    2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,
    2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,
    2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,
    2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,
    2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,
    2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,
    2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,
    2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,
    2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,
    2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,
    2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,
    2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,
    2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,
    2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,
    2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,
    4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
    4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
    4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
    4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
    4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
    4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
    4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
    4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
    4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
    4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
    4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
    4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
    4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
    4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
    4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
    4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
};

/*  This is the list of applications resources. These declarations are
    used by the toolkit to translate resource file specifications into
    values in the application resource data structure. */
static XtResource app_resource_list [] =
{
    {
	MslideNbufferPixmap,
	MslideCBufferPixmap,
	XtRBoolean,
	sizeof (Boolean),
	XtOffsetOf (application_resource_type, use_buffer_pixmap),
	XtRImmediate,
	(XtPointer) False
    },
    {
	MslideNclearBackground,
	MslideCClearBackground,
	XtRBoolean,
	sizeof (Boolean),
	XtOffsetOf (application_resource_type, clear_bkgnd),
	XtRImmediate,
	(XtPointer) False
    },
    {
	MslideNcolorNumColumns,
	MslideCColorNumColumns,
	XtRInt,
	sizeof (int),
	XtOffsetOf (application_resource_type, color_num_cols),
	XtRImmediate,
	(XtPointer) COLOR_NUM_COLS_DEFAULT
    },
    {
	MslideNdebugFlag,
	MslideCDebugFlag,
	XtRBoolean,
	sizeof (Boolean),
	XtOffsetOf (application_resource_type, debug),
	XtRImmediate,
	(XtPointer) False
    },
    {
	MslideNpaletteSize,
	MslideCPaletteSize,
	XtRInt,
	sizeof (int),
	XtOffsetOf (application_resource_type, palette_size),
	XtRImmediate,
	(XtPointer) PALETTE_SIZE_DEFAULT
    },
};

static application_resource_type    app_resources;

/*  This table describes the command line options we want the toolkit
    to process and the resource specifications that should result. */
static XrmOptionDescRec		    app_options [] =
{
    {"-buf_pixmap", "*bufferPixmap",	XrmoptionNoArg,	    "True"},
    {"-clear_bkgnd","*clearBackground",	XrmoptionNoArg,	    "True"},
    {"-color_cols", "*colorNumColumns",	XrmoptionSepArg,    NULL  },
    {"-debug",	    "*debugFlag",	XrmoptionNoArg,	    "True"},
    {"-palette_sz", "*paletteSize",	XrmoptionSepArg,    NULL  },
};

/*
**
**  CALLBACK ROUTINE DECLARATIONS
**
*/

static void cb_center (Widget w,
		       String tag_string,
		       XmAnyCallbackStruct *callback_data);
static void cb_create (Widget w,
		       String tag_string,
		       XmAnyCallbackStruct *callback_data);
static void cb_destroy (Widget w,
		        int *tag,
		        XmAnyCallbackStruct *callback_data);
static void cb_font_search (Widget w,
			    String tag_string,
			    XmAnyCallbackStruct *callback_data);
static void cb_help (Widget w,
		     String tag_string,
		     XmAnyCallbackStruct *callback_data);
static void cb_master_edit (Widget w,
			    String tag_string,
			    XmAnyCallbackStruct *callback_data);
static void cb_menu_choice (Widget w,
			    String tag_string,
			    XmAnyCallbackStruct *callback_data);
static void cb_menu_master (Widget w,
			    int *tag,
			    XmAnyCallbackStruct *callback_data);
static void cb_palette_edit (Widget w,
			     String tag_string,
			     XmAnyCallbackStruct *callback_data);
static void cb_slide_display (Widget w,
			      String tag_string,
			      XmAnyCallbackStruct *callback_data);
static void cb_slide_draw (Widget w,
			   String tag_string,
			   XmDrawingAreaCallbackStruct *draw_cb);
static void cb_slide_edit (Widget w,
			   String tag_string,
			   XmAnyCallbackStruct *callback_data);
static void cb_txt_val (Widget w,
			String tag_string,
		        XmAnyCallbackStruct *cbs);

/*  Set up callback routine addresses for MRM registration. */
static MrmRegisterArg	reg_vec [] =
{
    {"cb_center",	    (caddr_t) cb_center},
    {"cb_create",	    (caddr_t) cb_create},
    {"cb_destroy",	    (caddr_t) cb_destroy},
    {"cb_font_search",	    (caddr_t) cb_font_search},
    {"cb_help",		    (caddr_t) cb_help},
    {"cb_master_edit",	    (caddr_t) cb_master_edit},
    {"cb_menu_choice",	    (caddr_t) cb_menu_choice},
    {"cb_menu_master",	    (caddr_t) cb_menu_master},
    {"cb_palette_edit",	    (caddr_t) cb_palette_edit},
    {"cb_slide_display",    (caddr_t) cb_slide_display},
    {"cb_slide_draw",	    (caddr_t) cb_slide_draw},
    {"cb_slide_edit",	    (caddr_t) cb_slide_edit},
    {"cb_txt_val",	    (caddr_t) cb_txt_val},
};
static MrmCount		reg_num = XtNumber (reg_vec);

/*
**
**  FUNCTION PROTOTYPE DECLARATIONS
**
*/

static void add_menu_button (Widget w,
			     String lbl);
static Widget add_menu_get_item (Widget w);
static void add_menu_pixmap (Widget w, Pixmap pix);
static Boolean alloc_color_cell (Widget w,Pixel *pixp);
static void begin_rename_master (void);
static Widget build_paledit_dialog (palette_type *pal_ptr);
static Boolean check_save_show (reclist_type *show,
				palette_type *color_palette,
				reclist_type *master_list,
				String tag_string);
static int check_screen_type (Widget w);
static void clear_element_entry (void);
static void clear_entry (void);
static void clear_list (Widget w);
static void clear_menu_buttons (Widget w);
static XmString cs_create_localized (String str1);
static String cs_first_segment (XmString cs1);
static void delete_item_from_list (Widget w, unsigned int sindx);
static void delete_widget_info (Widget w);
static void display_error_text (String msg_code);
static void do_menu_choice (String tag_string);
static void draw_slide
    (slide_info_type *slide, Widget w, Drawable drawable);
static void element_clear (master_element_type *ep);
static void element_copy
    (master_element_type *dst,master_element_type *src);
static master_element_type *element_create (void);
static void element_free (master_element_type *elem);
static XmString fetch_mrm_cs (String lit_name);
static int fetch_mrm_int (String lit_name);
static String fetch_mrm_string (String lit_name);
static slide_master_type *find_master (reclist_type *mlp,
				       String mastername,
				       Boolean create_flag);
static unsigned int find_master_index (reclist_type *mlp,
				       String mastername);
static String format_bit_mask (unsigned long int mval,
			       int nbits,
			       char c_0,
			       char c_1);
static void get_element_from_entry (master_element_type *elem);
static Boolean get_font_from_name (Widget w,
				   String fontname,
				   Font *fontp);
static Cardinal get_option_menu_index (Widget widopt);
static String get_option_menu_value (Widget widopt);
static Pixel get_pixel_from_color_num (palette_type *palp,
				       Cardinal cnum);
static Boolean get_pixel_from_name (Widget w,
				    String colorname,
				    Pixel *pixelp);
static void get_slide_from_entry (slide_info_type *slide);
static Widget get_widget_shell (Widget w);
static void handle_popup_menu
    (Widget	w,
     XtPointer	client_data,
     XEvent	*event,
     Boolean	*continue_to_dispatch);
static void install_image (String iname,
			   Screen *scrn,
			   int iformat,
			   unsigned int iwidth,
			   unsigned int iheight,
			   unsigned char *idata,
			   palette_type *pal_ptr);
static void load_list_from_master (slide_master_type *master);
static void load_list_from_show (reclist_type *show);
static unsigned int lookup_el_type (String el_str);
static void make_element_gc (Widget		    w,
			     master_element_type    *ep);
static void make_element_pixmap (Widget w,
				 master_element_type *ep);
static void master_copy (slide_master_type *dst,
			 slide_master_type *src);
static slide_master_type *master_create (void);
static void master_free (slide_master_type *master);
static void mrm_register_name (String regname, caddr_t regval);
static Boolean popup_dialog_box (String wid_name,
				 int widget_index,
				 Boolean manage_flag);
static void popup_file_dialog (XmString title_string);
static void popup_msg_dialog (String dialog_name, String body_text);
static void print_show (String filnam,
			reclist_type *show,
			reclist_type *master_list);
static void print_slide (FILE *out_file,
			 reclist_type *show,
			 unsigned int sindx,
			 reclist_type *master_list);
static void put_element_to_entry (master_element_type *elem);
static void put_item_to_list (Widget w,unsigned int sindx,String str);
static void put_items_to_list (Widget w,
			       unsigned int sindx,
			       String *str_list,
			       unsigned int str_count);
static void put_slide_to_entry (slide_info_type *slide);
static void put_text_string (Widget w,String str);
static void reclist_clear (reclist_type *rlist);
static reclist_type *reclist_create
    (reclist_create_fcn	    create_fcn,
     reclist_copy_fcn	    copy_fcn);
static void reclist_delete_item
    (reclist_type *rlist, unsigned int delindx);
static void *reclist_get_item_ptr
    (reclist_type *rlist, unsigned int rindx);
static void reclist_insert_item (reclist_type *rlist,
				 unsigned int insindx,
				 void *insrec);
static void reclist_replace_item (reclist_type *rlist,
				  unsigned int repindx,
				  void *reprec);
static void refresh_master_names (reclist_type *mlp);
static void replace_item_in_list
    (Widget w,unsigned int sindx,String str);
static Widget retrieve_widget_id (String widnam, int wididx);
static widget_entry_type *retrieve_widget_info (Widget w);
static void save_widget_info (String widnam,
			      int wididx,
			      Widget widid);
static void select_item_in_list (Widget w,unsigned int sindx);
static void set_cursor (Widget w, unsigned int cursor_shape);
static void set_disp_slide (unsigned int indx);
static void set_edit_state (unsigned int sel_indx,
			    unsigned int cnt,
			    char *prefix);
static void set_file_name (String filnam);

#ifdef APP_ICON_UIL
static void set_icon_uil (Widget w, String iconname);
#else
static void set_icon_bitmap (Widget w,
			     int iconwidth,
			     int iconheight,
			     char *icondata);
#endif

static void set_master_edit_name (String master_name);
static void set_option_menu_index (Widget widopt, Cardinal indx);
static void set_option_menu_value (Widget widopt,String strval);
static void setup_default_master (reclist_type *mlp,
				  String mastername,
				  master_element_type *srclist);
static void setup_default_master_list (reclist_type *mlp);
static palette_type *setup_palette (Widget w,
				    int psize);
static void setup_palette_colors (palette_type *pal_ptr,
				  String *cnames);
static void setup_popup_menu (String widname,
			      Widget parent);
static void slide_clear (slide_info_type *slide);
static void slide_copy (slide_info_type *dst,slide_info_type *src);
static slide_info_type *slide_create (void);
static void slide_free (slide_info_type *slide);
static Boolean store_color_in_cell (Screen *scrn,
				    String cname,
				    Pixel pix);
static void update_disp_slide (slide_info_type *slide);
static void update_paledit_info (palette_type *pal_ptr);
static void update_show_status (reclist_type *show);
static void update_slide_list_descr (unsigned int sel_indx,
				     unsigned int cnt);
static Boolean wp_print_show (XtPointer client_data);
static Boolean wp_read_show (XtPointer client_data);
static Boolean wp_write_show (XtPointer client_data);
static void write_color (FILE *out_file,
			 palette_type *pal_ptr,
			 unsigned int sindx);
static void write_master (FILE *out_file,
			  reclist_type *mlp,
			  unsigned int mindx);
static void write_show (String filnam,
			reclist_type *show,
			palette_type *palette,
			reclist_type *master_list);
static void write_slide (FILE *out_file,
			 reclist_type *show,
			 unsigned int sindx);

/**************************************************
**      Main line routine of maker program.
**************************************************/
main (int argc, char *argv[])
{
    Arg			args[10];
    Cardinal		ac, mrm_status;
    int			screen_visual_type;
    MrmType		widget_class;
    Widget		wid1;

    /*	Initialize the record list structure for the main
	slide show. Currently, only one slide show is
	supported, but the data structures are set up
	to support more than one at a time. */
    main_show = reclist_create
		    ((reclist_create_fcn) slide_create,
		     (reclist_copy_fcn) slide_copy);

    edit_master = master_create ();
    null_master = master_create ();

    main_master_list = reclist_create
		    ((reclist_create_fcn) master_create,
		     (reclist_copy_fcn) master_copy);

    disp_slide = slide_create ();
    disp_indx = 0;

    setup_default_master_list (main_master_list);

    /*	Initialize the Motif Resource Manager. */
    MrmInitialize ();

    /*	Create the application shell widget. As part of the process,
	the toolkit parses command line options, including those
	we have declared in "app_options". Also set the shell resource
	that permits the shell to resize after it is created;
	some child widget resize requests might otherwise be
	refused. */
    ac = 0;
    XtSetArg (args[ac],XmNallowShellResize,True); ac++;
    wid_top_level = XtAppInitialize (&app_context,
				     app_name,
				     app_options,
				     XtNumber (app_options),
    /* For X11R4, "(Cardinal *) &argc" may be required. */
				     &argc,
				     (String *) argv,
				     (String *) NULL,
				     args,
				     ac);

    /*	Get values for the application resources we have defined. */
    XtGetApplicationResources (wid_top_level,
			       &app_resources,
			       app_resource_list,
			       XtNumber (app_resource_list),
			       (ArgList) NULL,
			       (Cardinal) 0);

    main_display = XtDisplay (wid_top_level);

    /*	Set the debug flag from the application resource value. */
    debug_flag = app_resources.debug;

    /*	See what kind of screen we are running on; take an error exit
	if it is not acceptable. */
    screen_visual_type = check_screen_type (wid_top_level);
    if (
	(screen_visual_type != StaticColor) &&
	(screen_visual_type != PseudoColor)
       )
    {
    	printf ("Screen must be StaticColor or PseudoColor.\n");
	exit (0);
    };

    /*	Access the compiled UIL interface description. */
    mrm_status = MrmOpenHierarchy (uid_file_num,
				   uid_file_names,
				   (MrmOsOpenParamPtr *) NULL,
				   &s_MrmHierarchy);
    if (mrm_status != MrmSUCCESS)
    {
    	printf ("Can't open hierarchy.\n");
	exit (0);
    }

    /*	Register names for callback routines so widgets can be set up
	properly when they are created by the Motif Resource
	Manager. */
    mrm_status = MrmRegisterNames (reg_vec,reg_num);
    if (mrm_status != MrmSUCCESS)
    {
    	printf ("Can't register names.\n");
	exit (0);
    }

    /*	Register the value for the UIL "identifier" that specifies
	the number of columns in the color selection menus. */
    mrm_register_name ("prm_color_num_cols",
		       (caddr_t) app_resources.color_num_cols);

    /*	Fetch the main window widget and, with it, the rest of the
	main application window. */
    mrm_status = MrmFetchWidget (s_MrmHierarchy,
				 "main_window",
				 wid_top_level,
				 &wid_main_window,
				 &widget_class);
    if (mrm_status != MrmSUCCESS)
    {
    	printf ("Can't fetch main widget.\n");
	exit (0);
    }

    /*	Load the slide master names into the menus. */
    refresh_master_names (main_master_list);

    /*	Make the icon bar visible or not, depending on the state
	of the "icon bar" toggle button, which may be modified in
	the application defaults (resource) file. */
    if (XmToggleButtonGetState
	(retrieve_widget_id ("options_menu_icon_bar",0)))
    {
	XtManageChild (retrieve_widget_id ("main_icon_bar",0));
    }
    else
    {
	XtUnmanageChild (retrieve_widget_id ("main_icon_bar",0));
    }

    /*	Manage the main window and tell the toolkit to set up all the
	widgets in the main application window. */
    XtManageChild (wid_main_window);
    XtRealizeWidget (wid_top_level);

    /*	Get format strings for slide list description. */
    ks_slide_list_descr_sel =
	fetch_mrm_string ("ks_slide_list_descr_sel");
    ks_slide_list_descr_nosel =
	fetch_mrm_string ("ks_slide_list_descr_nosel");

    /*	Get colors for slide list description. */
    wid1 = retrieve_widget_id ("slide_list_descr",0);
    XtVaGetValues (wid1,
		   XmNbackground,   &slide_descr_pixel_nosel,
		   NULL);
    if (!get_pixel_from_name (wid1,
			      "Pink",
			      &slide_descr_pixel_sel))
    {
	slide_descr_pixel_sel = slide_descr_pixel_nosel;
    }

    /*	Set up the color palette with default colors. */
    main_palette = setup_palette (wid_main_window,
				  app_resources.palette_size);
    setup_palette_colors (main_palette,default_color_names);

    /*	Install a "logo" image that can be referenced by slide master
	graphic elements. */
    install_image (LOGO_IMAGE_NAME,
		   XtScreen (wid_main_window),
		   LOGO_IMAGE_FORMAT,
		   LOGO_IMAGE_WIDTH,
		   LOGO_IMAGE_HEIGHT,
		   logo_image_data,
		   main_palette);

    /*	Set up the current editing state. */
    update_show_status (main_show);
    set_file_name ("Untitled");

    /*	Now we set the application's icon. This can be done using a
	UIL declaration, or directly from bitmap data. To make a UIL
	bitmap (2-level pixmap), we must use MrmFetchBitmapLiteral,
	which is not present in some releases of Motif prior to
	Version 1.2.
    
	By default, we'll create a bitmap at run time, rather than
	fetching it from UIL. If the macro APP_ICON_UIL is defined,
	the UIL-declared icon will be used.

	We pass the widget ID of the application shell, since that is
	where the icon pixmap is specified. Here, we have only one top
	level shell, but a different icon could be set for additional
	separately iconifiable top level shells if we had any. */

#ifdef APP_ICON_UIL
    set_icon_uil (wid_top_level,"app_icon_bitmap");
#else
    set_icon_bitmap (wid_top_level,
		     APP_ICON_WIDTH,
		     APP_ICON_HEIGHT,
		     app_icon_data);
#endif

    /*	Set up the pop-up menu for the slide editing area. */
    setup_popup_menu ("pu_control_menu",
		      wid_main_window);

    /*	Pass control to the toolkit to start the application. */
    XtAppMainLoop (app_context);

    /* Should not be able to reach this return. */
    return (0);
}

/**************************************************
**      add_menu_button
**************************************************/
static void add_menu_button (Widget w,
			     String lbl)
{
    XmString		cs1;
    Widget		wid_item;

    /*	Get the next available menu item widget.
	Create a label from the specified
	string and put it to the widget. */
    wid_item = add_menu_get_item (w);
    cs1 = cs_create_localized (lbl);
    XtVaSetValues (wid_item,
		   XmNlabelType,	    XmSTRING,
		   XmNlabelString,	    cs1,
		   NULL);
    XmStringFree (cs1);
}

/**************************************************
**      add_menu_get_item
**************************************************/
static Widget add_menu_get_item (Widget w)
{
    Cardinal		wid_child_count;
    Widget		wid_menu, wid_item;
    WidgetList		wid_children;

    /*	Get the next available menu item in the submenu of the
	specified widget. If there is none, create a new one.
	The caller then sets up the new menu item with a text
	or pixmap label.
	First, get the widget ID of the menu linked to this widget. */
    XtVaGetValues (w,
		   XmNsubMenuId,	&wid_menu,
		   NULL);

    /* Get list of item widgets in menu. */
    XtVaGetValues (wid_menu,
		   XmNchildren,	    &wid_children,
		   XmNnumChildren,  &wid_child_count,
		   NULL);

    /*	Scan through item widgets, looking for an available
	(unmanaged) widget. Save its widget ID. */
    wid_item = (Widget) NULL;
    while ((!wid_item) && (wid_child_count))
    {
	if (!XtIsManaged (*wid_children))
	{
	    wid_item = *wid_children;
	}
	else
	{
	    wid_children++;
	    wid_child_count--;
	}
    }

    /*	If we didn't find an available menu item widget,
	create one now. */
    if (!wid_item)
    {
	wid_item = (Widget)
	    XmCreatePushButton (wid_menu,
				(String) NULL,
				(Arg *) NULL,
				(Cardinal) 0);
    }

    /*	Manage the menu item and return its widget ID. We could
	delay managing the widget until after the caller sets
	its label, but normally the parent menu is not visible
	when we are making changes in its content, so there is
	no apparent difference in doing it here. */
    XtManageChild (wid_item);
    return (wid_item);
}

/**************************************************
**      add_menu_pixmap
**************************************************/
static void add_menu_pixmap (Widget w, Pixmap pix)
{
    Cardinal		wid_child_count;
    Widget		wid_menu, wid_item;
    WidgetList		wid_children;

    /*	Get the next available menu item widget.
	Set the specified pixmap as its label. */
    wid_item = add_menu_get_item (w);
    XtVaSetValues (wid_item,
		   XmNlabelType,	XmPIXMAP,
		   XmNlabelPixmap,	pix,
		   NULL);
}

/**************************************************
**      alloc_color_cell
**************************************************/
static Boolean alloc_color_cell (Widget w,Pixel *pixp)
{
    Boolean		    ok;
    Colormap		    cmap1;
    Status		    ret_status;
    unsigned long	    dummy_plane_mask;

    cmap1 = XDefaultColormapOfScreen (XtScreen (w));
    ret_status = XAllocColorCells (XtDisplay (w),
				   cmap1,
				   False,
				   &dummy_plane_mask,
				   0,
				   pixp,
				   1);
    ok = (ret_status != 0);

    return (ok);
}

/**************************************************
**      begin_rename_master
**************************************************/
static void begin_rename_master (void)
{
    XmString		cs1;
    Widget		wid1, wid2;

    (void) popup_dialog_box ("master_rename_box", 0, False);

    wid1 = retrieve_widget_id ("master_rename_box",0);

    /*  Clear the prompt text value. Also, undo
	the "default button" status for the "OK"
	button, so the keyboard focus can be set to
	the text field instead of the "OK" button. */

    cs1 = cs_create_localized ("");
    XtVaSetValues (wid1,
		   XmNtextString,	cs1,
		   XmNdefaultButton,    (Widget) NULL,
		   NULL);
    XmStringFree (cs1);

    /*  Pop up the dialog. */
    XtManageChild (wid1);

    /*  Move keyboard focus to the text field. */
    wid2 = XmSelectionBoxGetChild
	    (wid1,XmDIALOG_TEXT);
    XmProcessTraversal (wid2,XmTRAVERSE_CURRENT);
}

/**************************************************
**      build_paledit_dialog
**************************************************/
static Widget build_paledit_dialog (palette_type *pal_ptr)
{
    Arg			args [10];
    Boolean		ok;
    Cardinal		ac, mrm_status;
    char		ch1 [20];
    Dimension		parent_width, parent_height, win_width,
			    win_height, clip_width, clip_height;
    MrmType		widget_class;
    Pixmap		pixmap1;
    unsigned int	ii;
    Widget		wid_dialog, wid_parent, wid_child, wid_clip,
			    wid1;
    XmString		cs1;

    /*  We should be building the color palette dialog for the first
	time; "popup_dialog_box" returns true if this is the case. */
    ok = popup_dialog_box ("palette_edit_dialog", 0, False);
    wid_dialog = (retrieve_widget_id ("palette_edit_dialog",0));

    if (ok)
    {
	/*  We must create the needed number of color entries.
	    First, get the widget ID of the XmRowColumn container
	    for the color entries. */
	wid_parent = (retrieve_widget_id ("paledit_colors_rc",0));

	/*  Now fetch the XmRowColumn containing the information for
	    one color. Do this once for each color in the palette.
	    To tell the instances apart, set the color index number
	    as the value of the XmNuserData resource.
	    When the individual widgets within this replicated
	    container are created, the value of this resource from
	    their ancestor can be retrieved in the creation callback
	    routine. */
	for (ii = 1;  (ok) && (ii <= pal_ptr->size);  ii++)
	{
	    ac = 0;
	    XtSetArg (args[ac],
		      XmNuserData, (XtPointer) ii); ac++;

	    /*	Initialize widget variable to NULL, since some (old)
		versions of Motif no-op the fetch otherwise. */
	    wid_child = (Widget) NULL;

	    mrm_status =
		MrmFetchWidgetOverride (s_MrmHierarchy,
					"paledit_color_rc",
					wid_parent,
					(String) NULL,
					args,
					ac,
					&wid_child,
					&widget_class);
	    ok = (mrm_status == MrmSUCCESS);

	    if (ok)
	    {
		/*  Manage the new color entry box, and set up
		    the static values of its children. */
		XtManageChild (wid_child);

		/*  Write the formatted color index number to a label
		    widget used as a "display only" text field. */
		wid1 = retrieve_widget_id ("paledit_color_num",ii);
		sprintf (ch1, "%4u.", ii);
		cs1 = cs_create_localized ((String) ch1);
		XtVaSetValues (wid1,
			       XmNlabelString,  cs1,
			       NULL);
		XmStringFree (cs1);

		/*  Set up a label widget to display a pixmap
		    with a sample of this entry's color. */
		wid1 = retrieve_widget_id
			("paledit_color_sample",ii);
		pixmap1 = pal_ptr->entries[ii - 1].sample_pixmap;
		XtVaSetValues (wid1,
			       XmNlabelPixmap,  pixmap1,
			       NULL);
	    }
	}
    }

    if (ok)
    {
	/*  Now we have the palette edit dialog box created, but the
	    scrolled window enclosing the row column with the palette
	    entries is not the right width. We can't use a fixed width
	    in the UIL declaration, because the necessary width may
	    vary with screen resolution, fonts, etc. With automatic
	    scrolling, we can't just let the scrolled window size
	    itself either, so we have to compute the right width
	    and set it ourselves.
	    First, we manage the dialog so the geometry of the row
	    column widget will be computed. Next, we get its width
	    and the width of the "clip window" of the scrolled window
	    widget. Then we increase the width of the scrolled window
	    by the amount necessary to make the clip window as wide
	    as the row column. */
	XtManageChild (wid_dialog);
	XtVaGetValues (wid_parent,
		       XmNwidth,	&parent_width,
		       XmNheight,	&parent_height,
		       NULL);

	wid1 = retrieve_widget_id ("paledit_colors_win",0);
	XtVaGetValues (wid1,
		       XmNwidth,	&win_width,
		       XmNheight,	&win_height,
		       XmNclipWindow,	&wid_clip,
		       NULL);
	XtVaGetValues (wid_clip,
		       XmNwidth,	&clip_width,
		       XmNheight,	&clip_height,
		       NULL);

	if (debug_flag)
	{
	    printf ("build_palette_dialog:\n");
	    printf ("   before adjustment:\n");
	    printf ("     row column      width = %u, height = %u\n",
		    parent_width, parent_height);
	    printf ("     scrolled window width = %u, height = %u\n",
		    win_width, win_height);
	    printf ("     clip window     width = %u, height = %u\n",
		    clip_width, clip_height);
	}

	/*  Adjust the scrolled window width. We throw in two extra
	    pixels to make sure we get no horizontal scroll bar. The
	    toolkit may do the calculation a little differently
	    depending on whether the initial width is too little or
	    too much, but this correction factor seems to work in
	    either case. */
	XtVaSetValues (wid1,
		       XmNwidth,	win_width + 2 +
					parent_width - 
					clip_width,
		       NULL);

	XtVaGetValues (wid1,
		       XmNwidth,	&win_width,
		       XmNheight,	&win_height,
		       NULL);
	XtVaGetValues (wid_clip,
		       XmNwidth,	&clip_width,
		       XmNheight,	&clip_height,
		       NULL);

	if (debug_flag)
	{
	    printf ("    after adjustment:\n");
	    printf ("     scrolled window width = %u, height = %u\n",
		    win_width, win_height);
	    printf ("     clip window     width = %u, height = %u\n",
		    clip_width, clip_height);
	}

	/*  Display the current color palette info. */
	update_paledit_info (pal_ptr);
    }

    return (ok ? wid_dialog : (Widget) NULL);
}

/**************************************************
**      cb_center
**************************************************/
static void cb_center (Widget w,
		       String tag_string,
		       XmAnyCallbackStruct *callback_data)
{
    Dimension		size1;
    int			offset_old, offset_new;

    if (strcmp ((char *) tag_string, "LEFT") == 0)
    {
    	XtVaGetValues (w,
		       XmNwidth,	&size1,
		       XmNleftOffset,	&offset_old,
		       NULL);
	offset_new = - (size1 / 2);
	if (offset_new != offset_old)
	{
	    XtVaSetValues (w,
			   XmNleftOffset,   offset_new,
			   NULL);
	}
    }
    else if (strcmp ((char *) tag_string, "TOP") == 0)
    {
    	XtVaGetValues (w,
		       XmNheight,	&size1,
		       XmNtopOffset,	&offset_old,
		       NULL);
	offset_new = - (size1 / 2);
	if (offset_new != offset_old)
	{
	    XtVaSetValues (w,
			   XmNtopOffset,    offset_new,
			   NULL);
	}
    }

    if (debug_flag)
    {
	printf ("Widget [%s] centered [%s]:\n",
		XtName (w), tag_string);
	printf ("size = %d, old offset = %d, new offset = %d\n",
		size1, offset_old, offset_new);
    }
}

/**************************************************
**      cb_create
**************************************************/
static void cb_create (Widget w,
		       String tag_string,
		       XmAnyCallbackStruct *callback_data)
{
    char		*indexp;
    int			wid_index;
    String		widget_name, wid_info_name;
    Widget		wid1;
    XtPointer		user_data;

    /*	This callback is invoked whenever we need to save the widget
	ID of a widget created by MRM. Normally, the UIL "tag_string"
	argument is an empty string; in this case, we associate the
	widget name with its ID, and set the "index" value to zero.

	In other cases, it may be more convenient to have several
	widgets with the same "reference name" that are distinguished
	by the "index" value. An example would be a set of text
	widgets that contain the same type of information; the "index"
	can then be used to find one of an "array" of similar widgets.
	The corresponding tag string syntax is "refname|nn", where
	"refname" is the desired reference name and "nn" is a
	formatted numeric value that is to be used as the index.

	In some other cases, we want to fetch a subtree of widgets
	more than once, and have the index value designate which
	group a particular widget belongs to. Here, the widget name
	is used as the reference name, but the index value is
	derived from the XmNuserData resource of an ancestor (the root
	of the widget subtree). */

    widget_name = XtName (w);
    wid_info_name = (String) NULL;
    wid_index = 0;
    switch (tag_string [0])
    {
    	case '>':
	    /*  Tag string specifies widget name and possibly
		an explicit index value. */

	    /*  Copy the tag string, skipping the first character. */
	    wid_info_name = XtNewString (&tag_string[1]);
	    /*  Look for the '|' separator. */
	    indexp = strrchr ((char *) wid_info_name, '|');
	    /*  If found, convert the index value and
		terminate the "reference name" portion. */
	    if (indexp)
	    {
		wid_index = atoi (indexp+1);
		*indexp = '\0';
	    }
	    /*  Save the widget reference name, index, and ID. */
	    save_widget_info (wid_info_name,wid_index,w);
	    break;
    	case '*':
	    /*	Derive the index value from the XmNuserData
		resource of an ancestor. Go up the widget tree
		until we find a non-zero value or until
		we run out of ancestors. */
	    wid1 = XtParent (w);
	    while ((wid_index == 0) && (wid1))
	    {
	    	XtVaGetValues (wid1,
			       XmNuserData, &user_data,
			       NULL);
		wid_index = (int) user_data;
		wid1 = XtParent (wid1);
	    }
	    /*	Fall into default case; no "break". */
    	default:
	    /*  If this is not a special case, save the
		widget name, index (zero), and ID. */
	    save_widget_info (widget_name,wid_index,w);
    }

    if (debug_flag)
    {
	printf ("Widget [%s] created:\n",
		widget_name);
	printf ("  name = [%s], index = %d, tag = [%s]\n",
		wid_info_name ? wid_info_name : widget_name,
		wid_index,
		tag_string ? tag_string : "<NULL>");
    }

    /*	If we allocated a copy of the tag string, free it. */
    XtFree (wid_info_name);
}

/**************************************************
**      cb_destroy
**************************************************/
static void cb_destroy (Widget w,
			int *tag,
			XmAnyCallbackStruct *callback_data)
{
    /*	This callback is invoked when a widget is destroyed. The
	corresponding information is removed from the widget
	database. */
    delete_widget_info (w);

    if (debug_flag)
    {
	printf ("Deleting widget [%s]\n", XtName (w));
    }
}

/**************************************************
**      cb_font_search
**************************************************/
static void cb_font_search (Widget w,
			    String tag_string,
			    XmAnyCallbackStruct *callback_data)
{
    Boolean		    done, sep;
    /*	The maximum length of an XLFD font name is 255 characters. */
    char		    font_name_str [256];
    char		    **font_name_list;
    int			    font_name_count, font_name_index;
    String		    str1, *str_ptr;
    XFontStruct		    *sample_font_struct;
    XmFontList		    sample_font_list;
    XmListCallbackStruct    *list_cb;
    /* Maximum XLFD font name length is 255 characters. */
    XmString		    cs1;
    XmStringCharSet	    font_tag;
    XmStringContext	    cxt;
    XmStringDirection	    dir;
    Widget		    wid_fsn_list, wid1;

    static String	    search_font_family = (String) NULL;
    static String	    search_font_weight = (String) NULL;
    static String	    search_font_slant = (String) NULL;
    static String	    search_font_size = (String) NULL;
    static String	    search_font_select_name = (String) NULL;

    done = False,
    list_cb = (XmListCallbackStruct *) callback_data;

    if (debug_flag)
    {
	printf ("cb_font_search: widget     [%s]\n",XtName (w));
	printf ("                tag_string [%s]\n",
		tag_string);
    }

    wid_fsn_list = retrieve_widget_id ("font_search_name_list",0);

    if (strcmp ((char *) tag_string, "OK") == 0)
    {
	/* Font search ok */
	if (search_font_select_name)
	{
	    put_text_string
		(retrieve_widget_id ("mentry_font",0),
		 search_font_select_name);
	}
	done = True;
    }
    else if (strcmp ((char *) tag_string, "CANCEL") == 0)
    {
	done = True;
    }

    else if (strcmp ((char *) tag_string, "SELECT") == 0)
    {
	/* Select font */
	if (debug_flag)
	{
	    printf ("Selected font list item = %d\n",
		    list_cb->item_position);
	}

	XtFree (search_font_select_name);
	search_font_select_name =
	    (String) cs_first_segment (list_cb->item);

	wid1 = retrieve_widget_id ("font_search_sample",0);
	XtVaGetValues (wid1,
		       XmNlabelString,	&cs1,
		       NULL);

	XmStringInitContext (&cxt,cs1);
	XmStringGetNextSegment (cxt,
				(char **) &str1,
				&font_tag,
				&dir,
				&sep);
	XmStringFreeContext (cxt);
	if (debug_flag)
	{
	    printf ("font_search_sample: label = %s, tag = %s\n",
		    str1,
		    font_tag);
	}
	XtFree (str1);
	XtFree (font_tag);
	XmStringFree (cs1);

	sample_font_struct =
	    XLoadQueryFont (XtDisplay (w),search_font_select_name);

	if (sample_font_struct)
	{
	    sample_font_list =
		XmFontListCreate (sample_font_struct,
				  "font_sample");
	}
	else
	{
	    sample_font_list = (XmFontList) NULL;
	}

	XtVaSetValues (wid1,
		       XmNfontList,	sample_font_list,
		       NULL);

	if (sample_font_list) XmFontListFree (sample_font_list);

	/*  It would seem logical to free the old XFontStruct here,
	    now that the font list entry using it has been freed.
	    In general, however, it may be dangerous to free the font
	    structure, since the toolkit caches font list data,
	    and there might be an outstanding reference to the font
	    we would be unloading. So, we'll just leave the font
	    structure (and the corresponding loaded font) to be
	    cleaned up when the application exits. */
    }

    else if (tag_string[0] == 'F')
    {
	/* A font characteristic specification. */
	str_ptr = (String *) NULL;
	switch (tag_string[1])
	{
	     case 'F':
		/* Font family */
		str_ptr = &search_font_family;
		break;
	     case 'W':
		/* Font weight */
		str_ptr = &search_font_weight;
		break;
	     case 'S':
		/* Font slant */
		str_ptr = &search_font_slant;
		break;
	     case 'P':
		/* Font point size (decipoints) */
		str_ptr = &search_font_size;
		break;
	}
	if (str_ptr)
	{
	    /*	A font characteristic was specified. Save it
		and create a new XLFD font name. */
	    XtFree (*str_ptr);
	    *str_ptr = (String)
		XtMalloc (strlen ((char *) tag_string) - 2 + 1);
	    strcpy ((char *) *str_ptr, (char *) tag_string + 2);

	    /* XLFD prefix and FOUNDRY field */
	    strcpy ((char *) font_name_str, "-*-");

	    /* FAMILY_NAME field */
	    if (search_font_family)
	    {
		strcat ((char *) font_name_str,
			search_font_family);
	    }
	    else
	    {
		strcat ((char *) font_name_str, "*");
	    }
	    strcat ((char *) font_name_str, "-");

	    /* WEIGHT_NAME field */
	    if (search_font_weight)
	    {
		strcat ((char *) font_name_str,
			search_font_weight);
	    }
	    else
	    {
		strcat ((char *) font_name_str, "*");
	    }
	    strcat ((char *) font_name_str, "-");

	    /* SLANT field */
	    if (search_font_slant)
	    {
		strcat ((char *) font_name_str,
			search_font_slant);
	    }
	    else
	    {
		strcat ((char *) font_name_str, "*");
	    }
	    strcat ((char *) font_name_str, "-");

	    /* SETWIDTH_NAME field */
	    strcat ((char *) font_name_str, "*-");

	    /* ADD_STYLE_NAME field */
	    strcat ((char *) font_name_str, "*-");

	    /* PIXEL_SIZE field */
	    strcat ((char *) font_name_str, "*-");

	    /* POINT_SIZE field */
	    if (search_font_size)
	    {
		strcat ((char *) font_name_str,
			search_font_size);
	    }
	    else
	    {
		strcat ((char *) font_name_str, "*");
	    }
	    strcat ((char *) font_name_str, "-");

	    /* RESOLUTION_X field */
	    strcat ((char *) font_name_str, "*-");

	    /* RESOLUTION_Y field */
	    strcat ((char *) font_name_str, "*-");

	    /* SPACING field */
	    strcat ((char *) font_name_str, "*-");

	    /* AVERAGE_WIDTH field */
	    strcat ((char *) font_name_str, "*-");

	    /* CHARSET_REGISTRY field */
	    strcat ((char *) font_name_str, "*-");

	    /* CHARSET_ENCODING field */
	    strcat ((char *) font_name_str, "*");

	    clear_list (wid_fsn_list);
	    put_item_to_list (wid_fsn_list,
			      0,
			      (String) font_name_str);

	    /*	Get a list of font names matching the XLFD font
		name pattern we generated. The third argument
		to XListFonts is an arbitrary limit on the number
		of names to be returned; it serves only to limit
		the amount of memory that may be allocated in
		fulfilling our request. */
	    font_name_list = XListFonts (main_display,
					 font_name_str,
					 500,
					 &font_name_count);

	    if (font_name_count)
	    {
		put_items_to_list (wid_fsn_list,
				   0,
				   (String *) font_name_list,
				   font_name_count);
		XFreeFontNames (font_name_list);
	    }

	    select_item_in_list (wid_fsn_list, 1);
	}
    }


    /*	Check to see if the font search dialog should be
	dismissed. */
    if (done)
    {
	/* Unmanage font search box */
	XtUnmanageChild
	    (retrieve_widget_id ("font_search_dialog",0));
    }

}

/**************************************************
**      cb_help
**************************************************/
static void cb_help (Widget w,
		     String tag_string,
		     XmAnyCallbackStruct *callback_data)
{
    Cardinal		mrm_status;
    Cursor		help_cursor;
    MrmType		widget_class;
    Widget		wid_help, wid_shell, wid1, wid2;
    widget_entry_type	*widinfo1;
    XmAnyCallbackStruct help_cbs;

    if (debug_flag)
    {
	printf ("cb_help: widget     [%s]\n", XtName (w));
	printf ("         tag_string [%s]\n", tag_string);
    }

    wid1 = retrieve_widget_id ("help_dialog",0);

    if (strcmp ((char *) tag_string, "%DISMISS") == 0)
    {
	/*  Dismiss the help dialog. */
    	if (wid1) XtUnmanageChild (wid1);
    }
    else if (strcmp ((char *) tag_string, "%VERSION") == 0)
    {
	/*  Display "about" dialog. */
	(void) popup_dialog_box ("help_version_dialog", 0, True);
    }
    else if (strcmp ((char *) tag_string, "%CONTEXT") == 0)
    {
	/*  Handle context-sensitive help. Call the toolkit to
	    have the user select a widget for which help is needed.
	    If a widget is selected, search up the widget tree for
	    a widget that has a help callback specified, and invoke
	    that callback. */
	help_cursor =
	    XCreateFontCursor (XtDisplay (w), XC_question_arrow);
	wid_help =
	    XmTrackingLocate (wid_top_level, help_cursor, False);
	help_cbs.reason = XmCR_HELP;
	help_cbs.event = callback_data->event;
	while (wid_help)
	{
	    if (XtHasCallbacks (wid_help, XmNhelpCallback) ==
		XtCallbackHasSome)
	    {
		XtCallCallbacks
		    (wid_help, XmNhelpCallback, &help_cbs);
	    	wid_help = (Widget) NULL;
	    }
	    else
	    {
	    	wid_help = XtParent (wid_help);
	    }
	}
	XFreeCursor (XtDisplay (w), help_cursor);
    }
    else
    {
	/*  Get the shell ancestor of the widget making the
	    help request. */
	wid_shell = get_widget_shell (w);

	/*  Display the help dialog. To avoid problems
	    with modal dialogs, we re-create the help
	    dialog if its new shell ancestor is different than
	    its old one. The requesting widget's shell
	    becomes the parent of the help dialog shell, which
	    in turn is the parent of the help dialog box. */
	if (wid1)
	{
	    if (get_widget_shell (XtParent (XtParent (wid1))) !=
		wid_shell)
	    {
		XtDestroyWidget (wid1);
		wid1 = (Widget) NULL;
		delete_widget_info (wid1);
	    }
	}

	if (!wid1)
	{
	    mrm_status = MrmFetchWidget (s_MrmHierarchy,
					 "help_dialog",
					 wid_shell,
					 &wid1,
					 &widget_class);
	    if (mrm_status != MrmSUCCESS)
	    {
		printf ("Failed to fetch help dialog.\n");
		exit (0);
	    }
	    save_widget_info ("help_dialog", 0, wid1);
	}

	/*  If the help dialog is currently managed, unmanage it.
	    We do this in case the window manager has unmapped
	    the help dialog shell's window because the dialog
	    was the child of another transient shell. In this
	    case, the help dialog may be managed but unmapped,
	    confusing the toolkit. */
	if (XtIsManaged (wid1)) XtUnmanageChild (wid1);

	/*  Now load the appropriate help message into the help text
	    widget. The callback tag string is the message string
	    declared in UIL. */
	wid2 = retrieve_widget_id ("help_text",0);
	if (wid2)
	{
	    put_text_string (wid2, tag_string);

	    /*  Pop up the dialog. */
	    XtManageChild (wid1);
	}
    }
}

/**************************************************
**      cb_master_edit
**************************************************/
static void cb_master_edit (Widget w,
			    String tag_string,
			    XmAnyCallbackStruct *callback_data)
{
    Boolean		    done, master_changed;
    int			    st;
    master_element_type	    *temp_element, *ep;
    slide_master_type	    *mp;
    String		    str1;
    unsigned int	    element_index, master_index;
    XmListCallbackStruct    *list_cb;
    XmSelectionBoxCallbackStruct    *sel_cb;
    XmString		    cs1, cs2;
    Widget		    wid_elem_list, wid1, wid2;

    done = False;
    master_changed = False;
    wid_elem_list = retrieve_widget_id ("element_list",0);

    temp_element = element_create ();

    if (debug_flag)
    {
	printf ("cb_master_edit: widget     [%s]\n",
		XtName (w));
	printf ("                tag_string [%s]\n",
		tag_string);
    }

    if (strcmp ((char *) tag_string, "CLEAR") == 0)
    {
	/* Clear entry area */
	clear_element_entry ();
    }

    else if (strncmp ((char *) tag_string, "GET_", 4) == 0)
    {
	element_index = 0;
	if (strcmp ((char *) tag_string, "GET_PREV") == 0)
	{
	     /* Get previous element */
	     element_index =
		MAX (edit_master->elem_list->select_index - 1, 0);
	}
	else if (strcmp ((char *) tag_string, "GET_SELECT") == 0)
	{
	     /* Get selected element */
	     element_index = edit_master->elem_list->select_index;
	}
	else if (strcmp ((char *) tag_string, "GET_NEXT") == 0)
	{
	     /* Get next element */
	     if (edit_master->elem_list->select_index <
		 edit_master->elem_list->count)
	     {
		 element_index =
		    edit_master->elem_list->select_index + 1;
	     }
	}

	/*  If we have a specified element, put it into the
	    editing area, and make sure it is selected in
	    the list. */
	if (element_index)
	{
	    ep = (master_element_type *)
		reclist_get_item_ptr (edit_master->elem_list,
				      element_index);
	    put_element_to_entry (ep);
	    if (element_index != edit_master->elem_list->select_index)
	    {
		select_item_in_list (wid_elem_list, element_index);
	    }
	}
    }

    else if (strncmp ((char *) tag_string, "INSERT_", 7) == 0)
    {
	element_index = 0;
	if (strcmp ((char *) tag_string, "INSERT_BEFORE") == 0)
	{
	    /* Insert before selected element */
	    element_index = (edit_master->elem_list->select_index) ?
			edit_master->elem_list->select_index : 1;
	}
	else if (strcmp ((char *) tag_string, "INSERT_AFTER") == 0)
	{
	    /* Insert after selected element */
	    element_index = (edit_master->elem_list->select_index) ?
			edit_master->elem_list->select_index + 1 : 0;
	}

	/*  Get the element data from the entry area and insert it
	    into the slide master and the visible list. */
	get_element_from_entry (temp_element);
	reclist_insert_item (edit_master->elem_list,
			     element_index,
			     (void *) temp_element);
	put_item_to_list
	    (wid_elem_list,
	     element_index,
	     element_type_names [temp_element->element_type]);
	select_item_in_list (wid_elem_list,element_index);
    }

    else if (strcmp ((char *) tag_string, "REPLACE") == 0)
    {
	/* Replace selected element */
	 element_index = edit_master->elem_list->select_index;
	if ((element_index> 0) &&
	    (element_index <= edit_master->elem_list->count))
	{
	    get_element_from_entry (temp_element);
	    reclist_replace_item (edit_master->elem_list,
				  element_index,
				  (void *) temp_element);
	    replace_item_in_list
		(wid_elem_list,
		 element_index,
		 element_type_names [temp_element->element_type]);
	    select_item_in_list (wid_elem_list,
				 element_index);
	}
    }

    else if (strcmp ((char *) tag_string, "DELETE") == 0)
    {
	/* Delete selected element */
	element_index = edit_master->elem_list->select_index;
	if (element_index)
	{
	    reclist_delete_item (edit_master->elem_list,
				 element_index);
	    delete_item_from_list (wid_elem_list, element_index);
	    set_edit_state (edit_master->elem_list->select_index,
			    edit_master->elem_list->count,
			    "mctrl_");
	}
    }

    else if (strcmp ((char *) tag_string, "SELECT") == 0)
    {
	/*  Select element.
	    Cast callback structure pointer to proper type. */
	list_cb = (XmListCallbackStruct *) callback_data;
	if (debug_flag)
	{
	    printf ("Selected element list item = %d\n",
		    list_cb->item_position);
	}
	edit_master->elem_list->select_index = list_cb->item_position;
	set_edit_state (edit_master->elem_list->select_index,
			edit_master->elem_list->count,
			"mctrl_");
    }

    else if (strcmp ((char *) tag_string, "FONT_SEARCH") == 0)
    {
	/* Search for font -- pop up dialog. */
	get_element_from_entry (temp_element);
	(void) popup_dialog_box ("font_search_dialog", 0, True);
    }

    else if ((strcmp ((char *) tag_string, "MASTER_OK") == 0) ||
	     (strcmp ((char *) tag_string, "MASTER_APPLY") == 0))
    {
	/*  Store modified master. If "OK", we are done; if "Apply",
	    continue editing. */
	done = (strcmp ((char *) tag_string, "MASTER_OK") == 0);

	mp = find_master (main_master_list,edit_master->name,True);
	master_copy (mp,edit_master);
	main_master_list->modified_flag = True;
	refresh_master_names (main_master_list);
	master_changed = True;

	/*  In case this was "Apply", we must update the sensitivity
	    of the "Delete Master" button. Since we know we now have a
	    master with the current name, the only question is whether
	    we have only one (if so, can't delete it). */
	XtSetSensitive (retrieve_widget_id ("master_delete",0),
			(main_master_list->count > 1));
    }
    else if (strcmp ((char *) tag_string, "MASTER_CANCEL") == 0)
    {
	/* Master edit canceled */
	done = True;
    }

    else if (strcmp ((char *) tag_string, "MASTER_DELETE") == 0)
    {
	/*  Delete master; can't delete last one. */
	if (main_master_list->count > 1)
	{
	    popup_msg_dialog ("master_del_conf", edit_master->name);
	}
	else
	{
	    /*	Shouldn't be able to get here because the "Delete
		master" button should be insensitive. */
	}
    }

    else if (strcmp ((char *) tag_string, "MASTER_DELETE_OK") == 0)
    {
	/*  Delete master; can't delete last one. */
	master_index = find_master_index (main_master_list,
					  edit_master->name);
	if ((master_index) &&
	    (main_master_list->count > 1))
	{
	    reclist_delete_item (main_master_list,
				 master_index);
	    refresh_master_names (main_master_list);
	    master_changed = True;
	    done = True;
	}
    }

    else if (strcmp ((char *) tag_string, "MASTER_RENAME") == 0)
    {
	/* Begin rename request */
	begin_rename_master ();
    }
    else if (strcmp ((char *) tag_string, "MASTER_RENAME_OK") == 0)
    {
	/*  Rename OK. Cast callback struct pointer to proper type,
	    get new master name. */
	sel_cb = (XmSelectionBoxCallbackStruct *) callback_data;
	str1 = cs_first_segment (sel_cb->value);
	set_master_edit_name (str1);
	XtFree (str1);
    }
    else if (strcmp ((char *) tag_string,
		     "MASTER_RENAME_CANCEL") == 0)
    {
	/* Rename canceled -- no action. */
    }

    if (master_changed)
    {
	/*  If we modified a master and the slide form is up,
	    redraw it in case the current slide uses the master. */
	wid1 = retrieve_widget_id ("slide_form",0);
	if ((wid1) && (XtIsManaged (wid1)))
	{
	    update_disp_slide (disp_slide);
	}
    }

    if (done)
    {
	/* Unmanage master edit box; re-enable menu choice. */
	XtUnmanageChild (retrieve_widget_id ("master_form",0));
	XtSetSensitive
	    (retrieve_widget_id ("edit_menu_edit_master",0),True);
    }

    element_free (temp_element);
}

/**************************************************
**	cb_menu_choice
**************************************************/
static void cb_menu_choice (Widget w,
			    String tag_string,
			    XmAnyCallbackStruct *callback_data)
{
    FILE		    *test_file;
    String		    str1, str2;
    struct read_show_info   *wpd_ptr;
    Widget		    wid_check_save, wid_file_sel, wid1;
    XmFileSelectionBoxCallbackStruct	*fs_cbstruct;
    XmToggleButtonCallbackStruct	*tg_cbstruct;
    XmString		    cs1, cs_open, cs_save_as;

    wid_check_save = retrieve_widget_id ("check_save_box",0);
    wid_file_sel = retrieve_widget_id ("file_selection_box",0);

    if (debug_flag)
    {
	printf ("cb_menu_choice: widget     [%s]\n",XtName (w));
	printf ("                tag_string [%s]\n",
		tag_string);
    }

    if (
	(strcmp ((char *) tag_string, "EXIT") == 0) ||
	(strcmp ((char *) tag_string, "NEW") == 0) ||
	(strcmp ((char *) tag_string, "OPEN") == 0)
       )
    {
    /*  These menu choices have the effect of destroying
	the current slide show.
	For this reason, action on these requests may be
	delayed until the user decides whether to save
	unsaved changes to the current show. */
	if (check_save_show (main_show,
			     main_palette,
			     main_master_list,
			     tag_string))
	{
	    do_menu_choice (tag_string);
	}
    }
    else if (strncmp ((char *) tag_string, "SAVE", 4) == 0)
    {
	/* Save (or save as) slide show */
	do_menu_choice (tag_string);
    }
    else if (strcmp ((char *) tag_string, "CHECK_SAVE_YES") == 0)
    {
	/*  "Check save" box activation - "yes".
	    Save show, then do pending operation. */
	do_menu_choice ("SAVE");
	do_menu_choice (check_save_tag);
	XtUnmanageChild (wid_check_save);
    }
    else if (strcmp ((char *) tag_string, "CHECK_SAVE_NO") == 0)
    {
	/*  "Check save" box activation - "no".
	    Don't save show, but do pending operation. */
	do_menu_choice (check_save_tag);
	XtUnmanageChild (wid_check_save);
    }
    else if (strcmp ((char *) tag_string, "CHECK_SAVE_CANCEL") == 0)
    {
	/*  "Check save" box activation - "cancel".
	    Don't save show, discard pending operation. */
	XtUnmanageChild (wid_check_save);
    }

    else if (strcmp (tag_string, "FILE_OK") == 0)
    {
	/*  User selected "OK" from file selection dialog.
	    Pick up the file name from the callback data
	    structure.
	    Check the dialog title to find out whether this
	    is an "open" or a "save as" operation. */
	fs_cbstruct =
	    (XmFileSelectionBoxCallbackStruct *) callback_data;
	str1 = cs_first_segment (fs_cbstruct->value);

	XtVaGetValues (w,
		       XmNdialogTitle,	&cs1,
		       NULL);
	cs_open = fetch_mrm_cs("kcs_open_file_title");
	cs_save_as = fetch_mrm_cs("kcs_save_as_file_title");

	if (XmStringCompare (cs1, cs_open))
	{
	    /* OK on "open slide show" file selection. */
	    if (debug_flag)
		printf ("OK on file box - OPEN\n");

	    /*	Allocate and initialize data structure for
		work procedure. */
	    wpd_ptr = XtNew (struct read_show_info);

	    wpd_ptr->filename = XtNewString (str1);
	    wpd_ptr->show_ptr = main_show;
	    wpd_ptr->color_palette_ptr = main_palette;
	    wpd_ptr->master_list_ptr = main_master_list;
	    wpd_ptr->filep = (FILE *) NULL;
	    wpd_ptr->new_slide = (slide_info_type *) NULL;
	    wpd_ptr->new_master = (slide_master_type *) NULL;
	    wpd_ptr->new_element = (master_element_type *) NULL;

	    /*	Request work procedure. */
	    (void) XtAppAddWorkProc
			(app_context,
			 (XtWorkProc) wp_read_show,
			 (XtPointer) wpd_ptr);

	    /*	Pop down the file selection dialog. */
	    XtUnmanageChild (wid_file_sel);
	}
	else if (XmStringCompare (cs1, cs_save_as))
	{
	    /* OK on "save slide show as" file selection. */
	    if (debug_flag)
		printf ("OK on file box - SAVE AS\n");
#ifdef VMS
	    /*	On VMS, strip the version number from the file
		name. */
	    str2 = strchr (str1, ';');
	    if (str2) *str2 = '\0';
#endif
	    /*	Now we have the file name to save the slide show.
		Check to see if there is an existing file with the
		same name. */
	    test_file = fopen (str1, "r");
	    if (test_file)
	    {
		/*  If there is an existing file, close it and put
		    up a dialog to ask what to do. */
		fclose (test_file);
		XtFree (overwrite_file_name);
		overwrite_file_name = XtNewString (str1);
		popup_msg_dialog ("overwrite_conf",
				  overwrite_file_name);
	    }
	    else
	    {
		/*  If there is no such file, write out the show and
		    pop down the file selection dialog. */
		write_show
		    (str1,main_show,main_palette,main_master_list);
		XtUnmanageChild (wid_file_sel);
	    }
	}

	XmStringFree (cs1);
	XmStringFree (cs_open);
	XmStringFree (cs_save_as);
	XtFree (str1);
    }
    else if (strcmp ((char *) tag_string, "FILE_CANCEL") == 0)
    {
	/*  User selected "Cancel" from file selection dialog. */
	XtUnmanageChild (wid_file_sel);
    }

    else if (strcmp ((char *) tag_string, "FILE_OVERWRITE_OK") == 0)
    {
	/*  User selected "OK" from file overwrite dialog.
	    Start writing the show, and pop down the file selection
	    dialog. The message dialog is set to unmanage itself. */
	write_show (overwrite_file_name,
		    main_show,
		    main_palette,
		    main_master_list);
	XtUnmanageChild (wid_file_sel);
    }

    else if (strcmp ((char *) tag_string, "PRINT") == 0)
    {
	/*  Print show. For now, the print file name is fixed. */
	print_show ("mslide.lis", main_show, main_master_list);
    }
    else if (strcmp ((char *) tag_string, "ICON_BAR_STATE") == 0)
    {
	/*  The state of the "icon bar" toggle button changed.
	    Make the icon bar visible or invisible, as
	    requested. */
	tg_cbstruct = (XmToggleButtonCallbackStruct *) callback_data;
	wid1 = retrieve_widget_id ("main_icon_bar",0);

	if (tg_cbstruct->set)
	{
	    XtManageChild (wid1);
	}
	else
	{
	    XtUnmanageChild (wid1);
	}
    }

    /*	Update various menu options to reflect slide show state. */
    update_show_status (main_show);
}

/**************************************************
**      cb_menu_master
**************************************************/
static void cb_menu_master (Widget w,
			    int *tag,
			    XmAnyCallbackStruct *callback_data)
{
    Boolean		    new_fetch;
    char		    ch1 [20];
    palette_type	    *pal_ptr;
    Pixmap		    pixmap1;
    slide_master_type	    *mp;
    String		    str1;
    unsigned int	    indx;
    Widget		    wid1, wid2;
    XmString		    cs1;
    XmRowColumnCallbackStruct	*rc_cbstruct;

    pal_ptr = main_palette;
    rc_cbstruct = (XmRowColumnCallbackStruct *) callback_data;

    if (debug_flag)
    {
	printf ("cb_menu_master: widget     [%s]\n",XtName (w));
	printf ("                tag_int    [%d]\n",*tag);
    }

    XtVaGetValues (rc_cbstruct->widget,
		   XmNlabelString,  &cs1,
		   NULL);
    str1 = cs_first_segment (cs1);
    XmStringFree (cs1);

    new_fetch = popup_dialog_box ("master_form", 0, False);

    /*  Disable the "edit master" menu choice so the user can't
	start a new master edit without finishing this one. */
    XtSetSensitive
	(retrieve_widget_id ("edit_menu_edit_master",0),False);

    if (new_fetch)
    {
	/*  Set up the element type names on the option menu. This
	    must be done after the "master_form" widgets are fetched.
	    It need be done only the first time the form is popped
	    up. */
	wid1 = retrieve_widget_id ("mentry_type",0);
	clear_menu_buttons (wid1);
	for (indx = EL_TYPE_MIN;  indx <= EL_TYPE_MAX;  indx++)
	{
	    add_menu_button (wid1,element_type_names [indx]);
	}

	/*  Set up the color index numbers on the option menu,
	    as we did for the element types. */
	wid1 = retrieve_widget_id ("mentry_fgcolor",0);
	wid2 = retrieve_widget_id ("mentry_bgcolor",0);
	clear_menu_buttons (wid1);
	clear_menu_buttons (wid2);
	for (indx = 1;  indx <= pal_ptr->size;  indx++)
	{
	    /*  Add an entry for this palette entry to the foreground
		and background color option menus. If we have a sample
		color pixmap for this color, use it as a pixmap label
		for the option menu button. Otherwise, make a text
		label out of the color's index number. */
	    pixmap1 = pal_ptr->entries[indx - 1].sample_pixmap;
	    if (pixmap1 != XmUNSPECIFIED_PIXMAP)
	    {
		add_menu_pixmap (wid1,pixmap1);
		add_menu_pixmap (wid2,pixmap1);
	    }
	    else
	    {
		sprintf (ch1, "%u", indx);
		add_menu_button (wid1, (String) ch1);
		add_menu_button (wid2, (String) ch1);
	    }
	}

	/*  Set up the master editing pop-up menu. This only needs to
	    be done when the master edit form is first fetched; it
	    can't be done earlier since the parent widget doesn't
	    exist then. */
	setup_popup_menu
	    ("pu_mctrl_menu",
	      retrieve_widget_id ("master_border_box",0));
    }

    set_master_edit_name (str1);
    XtFree (str1);
    mp = find_master (main_master_list,edit_master->name,False);
    master_copy (edit_master,mp);
    load_list_from_master (edit_master);
    clear_element_entry ();
    edit_master->elem_list->modified_flag = False;
    edit_master->elem_list->select_index = 0;
    set_edit_state (edit_master->elem_list->select_index,
		    edit_master->elem_list->count,
		    "mctrl_");
    XtManageChild (retrieve_widget_id ("master_form",0));
}

/**************************************************
**      cb_palette_edit
**************************************************/
static void cb_palette_edit (Widget w,
			     String tag_string,
			     XmAnyCallbackStruct *callback_data)
{
    Boolean		done;
    palette_entry_type	*pe_ptr;
    palette_type	*pal_ptr;
    String		old_color_name;
    Widget		wid_dialog;
    widget_entry_type	*widinfo1;

    done = False;

    if (debug_flag)
    {
	printf ("cb_palette_edit: widget     [%s]\n",
		XtName (w));
	printf ("                 tag_string [%s]\n",
		tag_string);
    }

    pal_ptr = main_palette;
    wid_dialog = retrieve_widget_id ("palette_edit_dialog",0);

    if (strcmp ((char *) tag_string, "POPUP") == 0)
    {
	/*  Has the dialog already been created? */
	if (wid_dialog)
	{
	    /*  The dialog already exists; see if it is already
		managed (popped up). If so, the user may be
		requesting it again because it is obscured by another
		dialog (e.g., slide display). We pop it down (unmanage
		it) so it will be popped up on top of the other window
		when it comes back. In any case, pop up the dialog. */
	    if (XtIsManaged(wid_dialog)) XtUnmanageChild (wid_dialog);
	    XtManageChild (wid_dialog);
	}
	else
	{
	    /*	The dialog has not yet been created. Build it now
		and pop it up. */
	    wid_dialog = build_paledit_dialog (main_palette);
	    if (!wid_dialog)
	    {
		printf ("Can't build palette edit dialog.\n");
		exit (0);
	    }
	}

    }

    else if (strcmp ((char *) tag_string, "COLOR_CHANGE") == 0)
    {
	/*  The color name may have been modified. Compare the old and
	    new color names to see if the name has changed. If so,
	    mark the palette as modified and store the new pixel value
	    in the allocated cell (if we have one). */
	widinfo1 = retrieve_widget_info (w);
	pe_ptr = pal_ptr->entries + widinfo1->wid_index - 1;
	old_color_name = pe_ptr->color_name;
	pe_ptr->color_name = (String) XmTextGetString (w);
	if ((!old_color_name) ||
	    (!pe_ptr->color_name) ||
	    (strcmp ((char *) old_color_name,
		     (char *) pe_ptr->color_name) != 0))
	{
	    pal_ptr->modified_flag = True;
	    if (pe_ptr->allocated_pixel)
	    {
		(void) store_color_in_cell (pal_ptr->screen,
					    pe_ptr->color_name,
					    pe_ptr->pixel);
	    }
	}
	XtFree (old_color_name);
    }

    else if (strcmp ((char *) tag_string, "DISMISS") == 0)
    {
	/* Palette edit done */
	done = True;
    }

    if (done)
    {
	/* Unmanage palette edit dialog. */
	XtUnmanageChild (wid_dialog);
    }
}

/**************************************************
**      cb_slide_display
**************************************************/
static void cb_slide_display (Widget w,
			      String tag_string,
			      XmAnyCallbackStruct *callback_data)
{
    Boolean		    new_fetch;
    Cardinal		    wid_depth;
    Dimension		    wid_width, wid_height;
    Pixel		    bg_pix;
    String		    wid_name;
    unsigned int	    disp_indx2, draw_width, draw_height,
				draw_depth;
    Widget		    wid1;
    Window		    draw_root;
    XGCValues		    gc_values;

    if (debug_flag)
    {
	printf ("cb_slide_display: widget     [%s]\n", XtName (w));
	printf ("                  tag_string [%s]\n", tag_string);
    }

    disp_indx2 = 0;
    if (strcmp ((char *) tag_string, "DISPLAY") == 0)
    {
	/* Display slide */
	new_fetch = popup_dialog_box ("slide_form", 0, False);

	/*  If we are fetching the slide display dialog for the first
	    time, we must create a graphics context to clear the slide
	    display area. Since we use XFillRectangle to clear the
	    area, the background color of the widget is the foreground
	    color of the GC. */
	if (new_fetch)
	{
	    wid1 = retrieve_widget_id ("slide_area",0);
	    XtVaGetValues (wid1,
			   XmNbackground,	&bg_pix,
			   NULL);
	    gc_values.foreground = bg_pix;
	    draw_clear_gc = XtGetGC (w,
				     GCForeground,
				     &gc_values);

	    /*  If we are supposed to draw to a pixmap and then copy
		to the screen, allocate the pixmap now and clear it to
		the background color. We have to be careful here not
		to depend on the actual drawing area window, since the
		widget is not yet realized. */
	    if (app_resources.use_buffer_pixmap)
	    {
		XtVaGetValues (wid1,
			       XmNwidth,	    &wid_width,
			       XmNheight,	    &wid_height,
			       XmNdepth,	    &wid_depth,
			       NULL);
		draw_width = wid_width;
		draw_height = wid_height;
		draw_depth = wid_depth;
		draw_root = XRootWindowOfScreen (XtScreen (wid1));
		buffer_pixmap = XCreatePixmap (XtDisplay (wid1),
					       draw_root,
					       draw_width,
					       draw_height,
					       draw_depth);
		XFillRectangle (XtDisplay (wid1),
				buffer_pixmap,
				draw_clear_gc,
				0,		    0,
				draw_width + 1, draw_height + 1);
		if (debug_flag)
		{
		    printf ("                  ");
		    printf ("clear_gc and buffer pixmap\n");
		}
	    }
	}

	/*  Set up to display the selected slide, or the first slide
	    if none is selected. Unmanage the text widget that
	    displays the position of the last mouse click and pop up
	    the slide display dialog. */
	set_disp_slide (MAX(main_show->select_index,1));
	XtUnmanageChild (retrieve_widget_id ("slide_mouse_xy",0));
	XtManageChild (retrieve_widget_id ("slide_form",0));
    }
    else if (strcmp ((char *) tag_string, "FIRST") == 0)
    {
	/* First slide */
	disp_indx2 = 1;
    }
    else if (strcmp ((char *) tag_string, "PREV") == 0)
    {
	/* Previous slide */
	disp_indx2 = (disp_indx > 1) ?
			disp_indx - 1 : main_show->count;
    }
    else if (strcmp ((char *) tag_string, "NEXT") == 0)
    {
	/* Next slide */
	disp_indx2 = (disp_indx < main_show->count) ?
			 disp_indx + 1 : 1;
    }
    else if (strcmp ((char *) tag_string, "DISMISS") == 0)
    {
	/*  Dismiss slide display */
	XtUnmanageChild (retrieve_widget_id ("slide_form",0));
    }

    /* If a new slide was selected, display it. */
    if (disp_indx2) set_disp_slide (disp_indx2);
}

/**************************************************
**      cb_slide_draw
**************************************************/
static void cb_slide_draw (Widget w,
			   String tag_string,
			   XmDrawingAreaCallbackStruct *draw_cb)
{
    char		    ch1 [50];
    char		    key_string [1];
    Dimension		    wid_height, wid_width;
    int			    key_string_len;
    KeySym		    keysym;
    unsigned int	    disp_indx2;
    Widget		    wid1;
    XButtonEvent	    *mbev;
    XEvent		    *event;
    XExposeEvent	    *expev;

    if (debug_flag)
    {
	printf ("cb_slide_draw: widget     [%s]\n", XtName(w));
	printf ("               tag_string [%s]\n", tag_string);
    }

    /*	Get the event from the callback data structure. */
    event = draw_cb->event;

    if (strcmp ((char *) tag_string, "EXPOSE") == 0)
    {
	/* Expose callback: */
	expev = (XExposeEvent *) event;
	if (debug_flag)
	{
	    XtVaGetValues (w,
			   XmNheight,   &wid_height,
			   XmNwidth,    &wid_width,
			   NULL);

	    printf ("                      ");
	    printf ("expose - widget width,height = %u,%u\n",
		    wid_width,wid_height);
	    printf ("                      ");
	    printf ("expose - region x,y,width,height,count =");
	    printf (" %d,%d %dx%d %d\n",
		    expev->x,	    expev->y,
		    expev->width,   expev->height,
		    expev->count);
	}

	if (app_resources.use_buffer_pixmap)
	{
	    /*	We are drawing into a buffer pixmap and copying to
		the window when an exposure occurs, as it has now.
		Copy the exposed region from the pixmap to the
		window. */
	    XCopyArea (XtDisplay (w),
		       buffer_pixmap,
		       XtWindow (w),
		       XDefaultGCOfScreen (XtScreen (w)),
		       expev->x,	expev->y,
		       (unsigned int) expev->width,
		       (unsigned int) expev->height,
		       expev->x,	expev->y);
	}
	else
	{
	    /*  We are drawing directly to the window. */

	    /*	See if there are any known pending exposure events
		still on the way (count != 0); if not, redraw the
		slide now.
		If more events are coming, ignore all but the last in
		the sequence. We do this because the draw_slide
		routine always redraws the whole slide image, so
		there is no benefit to processing each exposed region
		separately. */
	    if (expev->count == 0)
		draw_slide (disp_slide, w, (Drawable) XtWindow (w));
	}

    }

    else if (strcmp ((char *) tag_string, "INPUT") == 0)
    {
    	/*  Input: keyboard or mouse button event. */
	switch (event->type)
	{

	    /*	For a mouse button 1 press, get the mouse pointer
		position, format it, and display it so the user can do
		some "manual" positioning of graphic elements.
		For any other button press, remove the text widget
		that holds the formatted position. */
	    case ButtonPress:
		mbev = (XButtonEvent *) event;
		wid1 = retrieve_widget_id ("slide_mouse_xy",0);
		if (mbev->button == Button1)
		{
		    sprintf (ch1, "x = %d, y=%d", mbev->x, mbev->y);
		    put_text_string (wid1,(String) ch1);
		    /*  Make sure the text widget is visible. */
		    XtManageChild (wid1);
		}
		else
		{
		    /*  Make text widget invisible. */
		    XtUnmanageChild (wid1);
		}

		/*  Set keyboard focus to the drawing area widget
		    so subsequent key events will be routed to it. */
	    	XmProcessTraversal (w, XmTRAVERSE_CURRENT);
		break;

	    /*	For a key press, convert the key code in the event
		data structure to a keysym. We look for certain keys
		to display the next slide, the previous slide, or
		the first slide in the show. */
	    case KeyPress:
		key_string_len =
		    XLookupString ((XKeyEvent *) event,
				   key_string,
				   sizeof key_string,
				   &keysym,
				   (XComposeStatus *) NULL);
		disp_indx2 = 0;
		switch (keysym)
		{
		    case XK_Home:
		    case XK_Up:
		    case XK_S:
		    case XK_s:
			disp_indx2 = 1;
			break;
		    case XK_Right:
		    case XK_F:
		    case XK_f:
			disp_indx2 = (disp_indx < main_show->count) ?
				     disp_indx + 1 : 1;
			break;
		    case XK_Left:
		    case XK_B:
		    case XK_b:
			disp_indx2 = (disp_indx > 1) ?
				     disp_indx - 1 : main_show->count;
			break;
		    default:
			XBell (XtDisplay (w), 0);
			if (debug_flag)
			{
			    printf ("keysym = %x, string = [%.*s]\n",
				    keysym,
				    key_string_len,
				    key_string);
			}
		}
		/* If a new slide was selected, display it. */
		if (disp_indx2) set_disp_slide (disp_indx2);
		break;

	/*  Note that we don't do any processing of button release or
	    key release events, though we could handle them here. */
	}
    }
}

/**************************************************
**      cb_slide_edit
**************************************************/
static void cb_slide_edit (Widget w,
			   String tag_string,
			   XmAnyCallbackStruct *callback_data)
{
    slide_info_type	    *temp_slide, *sp1;
    unsigned int	    slide_index;
    Widget		    wid_slide_list;
    XmListCallbackStruct    *list_cb;

    slide_index = 0;
    wid_slide_list = retrieve_widget_id ("slide_list",0);

    temp_slide = slide_create ();

    if (debug_flag)
    {
	printf ("cb_slide_edit: widget     [%s]\n", XtName (w));
	printf ("               tag_string [%s]\n",tag_string);
    }

    if (strcmp ((char *) tag_string, "CLEAR") == 0)
    {
	/* Clear entry area */
	clear_entry ();
    }
    else if (strncmp ((char *) tag_string, "GET_", 4) == 0)
    {
    	if (strcmp ((char *) tag_string, "GET_PREV") == 0)
	{
	    /* Get previous slide */
	    slide_index =
		MAX (main_show->select_index - 1, 0);
	}
    	else if (strcmp ((char *) tag_string, "GET_SELECT") == 0)
	{
	    /* Get selected slide */
	    slide_index = main_show->select_index;
	}
    	else if (strcmp ((char *) tag_string, "GET_NEXT") == 0)
	{
	    /* Get next slide */
	    if (main_show->select_index <
		main_show->count)
	    {
		slide_index = main_show->select_index + 1;
	    }
	}

	if (slide_index)
	{
	    sp1 = (slide_info_type *)
		reclist_get_item_ptr (main_show,slide_index);
	    put_slide_to_entry (sp1);
	    if (slide_index != main_show->select_index)
	    {
		select_item_in_list (wid_slide_list,
				     slide_index);
	    }
	}
    }
    else if (strncmp ((char *) tag_string, "INSERT_", 7) == 0)
    {
    	if (strcmp ((char *) tag_string, "INSERT_BEFORE") == 0)
	{
	    /* Insert before selected slide */
	    slide_index = (main_show->select_index) ?
			   main_show->select_index : 1;
	}
    	else if (strcmp ((char *) tag_string, "INSERT_AFTER") == 0)
	{
	    /* Insert after selected slide */
	    slide_index = (main_show->select_index) ?
			   main_show->select_index + 1 : 0;
	}

	get_slide_from_entry (temp_slide);
	reclist_insert_item (main_show,
			     slide_index,
			     (void *) temp_slide);
	put_item_to_list (wid_slide_list,
			  slide_index,
			  temp_slide->title_text);
	select_item_in_list (wid_slide_list,slide_index);
    }
    else if (strcmp ((char *) tag_string, "REPLACE") == 0)
    {
	/* Replace selected slide */
	 slide_index = main_show->select_index;
	if ((slide_index> 0) &&
	    (slide_index <= main_show->count))
	{
	    get_slide_from_entry (temp_slide);
	    reclist_replace_item (main_show,
				  slide_index,
				  (void *) temp_slide);
	    replace_item_in_list (wid_slide_list,
				  slide_index,
				  temp_slide->title_text);
	    select_item_in_list (wid_slide_list,
				 slide_index);
	}
    }
    else if (strcmp ((char *) tag_string, "DELETE") == 0)
    {
	/* Delete selected slide */
	slide_index = main_show->select_index;
	if (slide_index)
	{
	    reclist_delete_item (main_show,slide_index);
	    delete_item_from_list (wid_slide_list,
				   slide_index);
	}
    }
    else if (strcmp ((char *) tag_string, "SELECT") == 0)
    {
	/* Select slide. */
	list_cb = (XmListCallbackStruct *) callback_data;
	if (debug_flag)
	{
	    printf ("Selected list item = %d\n",
		    list_cb->item_position);
	}
	main_show->select_index = list_cb->item_position;
    }

    update_show_status (main_show);
    slide_free (temp_slide);
}

/**************************************************
**      cb_txt_val
**************************************************/
static void cb_txt_val (Widget w,
		        String tag_string,
		        XmAnyCallbackStruct *cbs)
{
    Boolean			ok;
    char			fmt_buff [20];
    int				int1,len;
    String			str1;
    XmTextVerifyCallbackStruct	*textv_cbs;
    XmTextBlock			blockp;

    if (debug_flag)
    {
	printf ("cb_txt_val: widget     [%s]\n", XtName (w));
	printf ("            tag_string [%s]\n", tag_string);
	printf ("            reason = %d\n", cbs->reason);
    }

    /*	Dispatch according to the type of callback, since the
	callback data structure differs among callback types.
	We handle only activate, losing focus, and modify verify
	callbacks. In text validation routines, there is the
	possibility of reentrancy if the callback modifies the
	text in such a way as to trigger another callback for the
	same widget. */
    switch (cbs->reason)
    {
	/*  Activate or losing focus: This validation is done when
	    the field entry is "complete". We don't use any data
	    from the callback structure (except the reason) here. */
    	case XmCR_ACTIVATE:
    	case XmCR_LOSING_FOCUS:
	    if (strcmp ((char *) tag_string, "FORMAT_NUM") == 0)
	    {
		/*  Format a number in standard form; e.g., remove
		    leading zeros, etc. Get the current text value,
		    format it, and put it back to the widget. If the
		    text string is empty, leave it alone. */
		str1 = (String) XmTextGetString (w);
		if ((str1) && (*str1))
		{
		    int1 = atoi (str1);
		    XtFree (str1);
		    sprintf (fmt_buff, "%d", int1);
		    /*	If the string value hasn't changed,
			put_text_string will not output it. */
		    put_text_string (w, (String) fmt_buff);
		}
	    }
	    else
	    {
		printf ("            bad validation type\n");
	    }
	    break;

	/*  Modify verify callback: called when a text modification
	    is pending. We inspect any text that is proposed for
	    insertion; if it is invalid, we reject the change by
	    setting the "doit" callback structure member to False.
	    For normal keyboard entry, we expect only one character
	    to be entered at a time, but the user could be copying
	    in text from elsewhere, resulting in more than one. */
    	case XmCR_MODIFYING_TEXT_VALUE:
	    ok = True;
	    textv_cbs = (XmTextVerifyCallbackStruct *) cbs;
	    blockp = textv_cbs->text;

	    if (strcmp ((char *) tag_string, "NUM") == 0)
	    {
		/*  Numeric digits only: check all characters
		    in the pending insertion. */
		str1 = blockp->ptr;
		len = blockp->length;
		while ((ok) && (len--))
		{
		    ok = isdigit (*str1++);
		}
	    }
	    else
	    {
		printf ("            bad validation type\n");
	    }

	    /*	Set validation result. If False, the toolkit
		will not make the attempted modification. */
	    textv_cbs->doit = ok;
	    break;

    	default:
	    printf ("            bad callback reason\n");
    }

}

/**************************************************
**      check_save_show
**************************************************/
static Boolean check_save_show (reclist_type *show,
				palette_type *color_palette,
				reclist_type *master_list,
				String tag_string)
{
    Boolean		process;
    Widget		wid1;

    /*	The user has requested an action that will discard the current
	slide show. If a slide master edit is in progress, display a
	message and ignore the request unless it is "EXIT". */
    process = True;
    wid1 = retrieve_widget_id ("master_form",0);
    if ((wid1) &&
	(XtIsManaged (wid1)) &&
	(strcmp ((char *) tag_string, "EXIT") != 0))
    {
	display_error_text ("kcs_warn_close_master");
	process = False;
    }

    /*	See if there are unsaved changes to the current slide
	show. If so, pop up a dialog box to ask what to do.
	The user response is handled by "cb_menu_choice". */
    if (process)
    {
	if ((show->modified_flag) ||
	    (color_palette->modified_flag) ||
	    (master_list->modified_flag))
	{
	    XtFree (check_save_tag);
	    check_save_tag = XtNewString (tag_string);
	    (void) popup_dialog_box ("check_save_box", 0, True);
	    process = False;

	    /*	Force keyboard focus to the "Yes" button. */
	    wid1 = retrieve_widget_id ("check_save_yes_button",0);
	    XmProcessTraversal (wid1, XmTRAVERSE_CURRENT);
	}
    }

    return (process);
}

/**************************************************
**      check_screen_type
**************************************************/
static int check_screen_type (Widget w)
{
    Boolean	default_visual_flag;
    int		visual_count, ii, ret_visual_type;
    long	visual_template_mask;
    String	str1 = (String) NULL;
    Visual	*def_visual;
    XVisualInfo	visual_template;
    XVisualInfo	*visual_list;

static char	*visual_type_names [] =
{
    "StaticGray",
    "GrayScale",
    "StaticColor",
    "PseudoColor",
    "TrueColor",
    "DirectColor"
};

    /*	Get the default visual for the screen of the specified
	widget. Assuming we have not explicitly changed visuals,
	this will be the visual we want to check.
	If the XmNvisual resource of the application shell is
	not "CopyFromParent" (the default), we would have to
	be more careful in finding the visual to check. */
    def_visual = XDefaultVisualOfScreen (XtScreen (w));

    /*	Set the desired visual characteristics that are used to
	select the visuals we want to look at. The only value
	we specify is the (default) visual ID. Note that the
	selection is also controlled by the mask value set
	in the next section. We really only need to set the
	"visualid" member; the others are done for completeness. */
    visual_template.visual = (Visual *) NULL;
    visual_template.visualid = XVisualIDFromVisual (def_visual);
    visual_template.screen = 0;
    visual_template.depth = 0;
    visual_template.class = 0;
    visual_template.red_mask = 0;
    visual_template.green_mask = 0;
    visual_template.blue_mask = 0;
    visual_template.colormap_size = 0;
    visual_template.bits_per_rgb = 0;

    /*	Create a bit mask to indicate which visual characteristics we
	are concerned about. For debugging, we set "VisualNoMask",
	meaning that even the visual ID we specified above will be
	ignored. Normally, we request only the visual matching the
	default visual ID, using the mask "VisualIDMask". */
    visual_template_mask = (debug_flag) ? VisualNoMask : VisualIDMask;

    /* Get the list of matching visuals. */
    visual_list = XGetVisualInfo (XtDisplay (wid_top_level),
				  visual_template_mask,
				  &visual_template,
				  &visual_count);

    /*	At this point, visual_list points to zero or more visual info
	structures; the number of matching visuals is in visual_count.
	In the normal case, we should have one (the default) visual;
	for debugging, we should have all visuals for this display. */

    for (ii = 0; ii < visual_count; ii++)
    {
	default_visual_flag =
	    (visual_list[ii].visualid == visual_template.visualid);

	if (default_visual_flag)
	{
	    ret_visual_type = visual_list[ii].class;
	}

	if (debug_flag)
	{
	    printf ("check_screen_type -- visual %d     %s\n",
		    (ii + 1),
		    default_visual_flag ? "(Default visual)" : "");
	    printf ("    screen:              %d\n",
		    visual_list[ii].screen);
	    printf ("    depth:               %u\n",
		    visual_list[ii].depth);
	    printf ("    type:                %s\n",
		    visual_type_names [visual_list[ii].class]);
	    switch (visual_list[ii].class)
	    {
	    	case StaticGray:
	    	case GrayScale:
	    	case StaticColor:
	    	case PseudoColor:
		    /* Color primary masks not meaningful. */
		    break;
	    	case TrueColor:
	    	case DirectColor:
		    printf ("    color masks:         ");
		    printf ("MSB                          LSB\n");
		    str1 = format_bit_mask
			    (visual_list[ii].red_mask,32,'-','X');
		    printf ("                   Red:  %s\n",str1);
		    XtFree ((char *) str1);
		    str1 = format_bit_mask
			    (visual_list[ii].green_mask,32,'-','X');
		    printf ("                 Green:  %s\n",str1);
		    XtFree ((char *) str1);
		    str1 = format_bit_mask
			    (visual_list[ii].blue_mask,32,'-','X');
		    printf ("                  Blue:  %s\n",str1);
		    XtFree ((char *) str1);
		    break;
	    }
	    printf ("    colormap size:       %d\n", 
		    visual_list[ii].colormap_size);
	    printf ("    color primary bits:  %d\n", 
		    visual_list[ii].bits_per_rgb);
	    printf ("\n");
	}
    }

    /*	Free the list of returned visuals. */
    /*	The type cast here should really be (void *), but some
	versions of Xlib.h require (char *). */
    if (visual_list) XFree ((char *) visual_list);

    /*	Return the visual type for the screen. Note that we could also
	check other characteristics, such as the colormap size. Most
	PseudoColor screens have a large enough colormap to run the
	application, but some PC-based servers with limited graphics
	(e.g., simple VGA) may not. */

    return (ret_visual_type);
}

/**************************************************
**      clear_element_entry
**************************************************/
static void clear_element_entry (void)
{
    set_option_menu_value
	(retrieve_widget_id ("mentry_type",0),"");

    put_text_string (retrieve_widget_id ("mentry_xpos",0),NULL);
    put_text_string (retrieve_widget_id ("mentry_ypos",0),NULL);
    put_text_string (retrieve_widget_id ("mentry_width",0),NULL);
    put_text_string (retrieve_widget_id ("mentry_height",0),NULL);
    put_text_string (retrieve_widget_id ("mentry_font",0),NULL);
    set_option_menu_index
	(retrieve_widget_id ("mentry_fgcolor",0),0);

    XmToggleButtonSetState
	(retrieve_widget_id ("elem_data_none",0),True,True);

    XmScaleSetValue (retrieve_widget_id ("mentry_data_index",0), 0);
}

/**************************************************
**      clear_entry
**************************************************/
static void clear_entry (void)
{
    unsigned int	indx;

    set_option_menu_value
	(retrieve_widget_id ("entry_master",0),"");

    put_text_string (retrieve_widget_id ("entry_header",0),NULL);

    XmToggleButtonSetState
	(retrieve_widget_id ("entry_bar",0),False,False);

    put_text_string (retrieve_widget_id ("entry_title",0),NULL);

    for (indx = 1;  indx <= SLIDE_BODY_MAX;  indx++)
    {
	XmToggleButtonSetState
	    (retrieve_widget_id ("entry_bullet", indx),
	     False,
	     False);
	put_text_string
	    (retrieve_widget_id ("entry_body",indx), NULL);
    }
}

/**************************************************
**      clear_list
**************************************************/
static void clear_list (Widget w)
{
/*---
    In Motif 1.1, XmListDeleteAllItems does not always clean up
    correctly. Specifically, the following sequence causes problems:

    Populate a large list.
    Position to the last item.
    Delete all items.
    Insert a smaller number of items using an
    insertion position of zero (at end of list).

    This causes a number of problems, including unspecified
    vertical resizing, etc.
    For now, use the work-around below instead of the following:

    XmListDeselectAllItems (w);
    XmListDeleteAllItems (w);
*/

    XtVaSetValues (w,
		   XmNitemCount,	    0,
		   XmNitems,		    (XmStringTable) NULL,
		   XmNselectedItemCount,    0,
		   XmNselectedItems,	    (XmStringTable) NULL,
		   NULL);
}

/**************************************************
**      clear_menu_buttons
**************************************************/
static void clear_menu_buttons (Widget w)
{
    Cardinal		wid_child_count;
    Widget		wid_menu;
    WidgetList		wid_children;

    /*	Get the widget ID of the menu linked to this widget. */
    XtVaGetValues (w,
		   XmNsubMenuId,	&wid_menu,
		   NULL);

    /*	Get list of item widgets in menu.
	They should all be push buttons, created by
	add_menu_button or add_menu_pixmap. */
    XtVaGetValues (wid_menu,
		   XmNchildren,	    &wid_children,
		   XmNnumChildren,  &wid_child_count,
		   NULL);

    /*	Unmanage all the menu items. */
    XtUnmanageChildren (wid_children,wid_child_count);
}

/**************************************************
**      cs_create_localized
**************************************************/
static XmString cs_create_localized (String str1)
{
    XmString		cs1;

    /*	Create a compound string in the current locale.
	In Motif Version 1.2+, we can use XmStringCreateLocalized;
	for compatibility with prior versions, XmStringCreate
	can be used with a default font tag.

    cs1 = XmStringCreateLocalized ((char *) str1);
    */

    cs1 = XmStringCreate ((char *) str1,XmSTRING_DEFAULT_CHARSET);

    return (cs1);
}

/**************************************************
**      cs_first_segment
**************************************************/
static String cs_first_segment (XmString cs1)
{
    Boolean		    sep;
    String		    str1;
    XmStringCharSet	    font_tag;
    XmStringContext	    cxt;
    XmStringDirection	    dir;

    XmStringInitContext (&cxt,cs1);
    XmStringGetNextSegment (cxt,
			    (char **) &str1,
			    &font_tag,
			    &dir,
			    &sep);
    XmStringFreeContext (cxt);
    XtFree (font_tag);

    return (str1);
}

/**************************************************
**      delete_item_from_list
**************************************************/
static void delete_item_from_list (Widget w, unsigned int sindx)
{
    int		item_cnt;

    item_cnt = 0;
    XtVaGetValues (w,
		   XmNitemCount,    &item_cnt,
		   NULL);

    if ((sindx > 0) && ((int) sindx <= item_cnt))
    {
	XmListDeletePos (w,sindx);
    }
    else
    {
	printf ("delete_item_from_list: ");
	printf (" attempt to delete item (%u) of %d\n",
		sindx, item_cnt);
    }

}

/**************************************************
**      delete_widget_info
**************************************************/
static void delete_widget_info (Widget w)
{
    widget_entry_type	*wp, *wp_found;
    widget_entry_type	**ptr_prev;

    wp = wid_list_head;
    ptr_prev = (widget_entry_type **) &wid_list_head;
    wp_found = (widget_entry_type *) NULL;

    while ((!wp_found) && (wp))
    {
	if (w == wp->wid_id)
	{
	    wp_found = wp;
	}
	else
	{
	    ptr_prev = &wp->next;
	    wp = wp->next;
	}
    }

    if (wp_found)
    {
    	*ptr_prev = wp_found->next;
	XtFree ((char *) wp_found);
    }
}

/**************************************************
**      display_error_text
**************************************************/
static void display_error_text (String msg_code)
{
    Arg			args[10];
    Cardinal		ac, mrm_stat;
    Widget		wid1, wid2;
    XmString		cs1;

    /*	Pop up the error dialog. */
    (void) popup_dialog_box ("error_dialog", 0, False);
    wid1 = retrieve_widget_id ("error_dialog",0);

    /*  Remove the HELP button. */
    wid2 = XmMessageBoxGetChild
	    (wid1,XmDIALOG_HELP_BUTTON);
    XtUnmanageChild (wid2);

    /*  Remove the CANCEL button. */
    wid2 = XmMessageBoxGetChild
	    (wid1,XmDIALOG_CANCEL_BUTTON);
    XtUnmanageChild (wid2);

    /*  Set the message string value using a value from
	UIL. If the message fetch fails, display the message
	code instead (better than nothing?). */
    ac = 0;
    XtSetArg (args[ac],XmNmessageString,msg_code); ac++;
    mrm_stat = MrmFetchSetValues (s_MrmHierarchy,
				  wid1,
				  args,
				  ac);
    if (mrm_stat != MrmSUCCESS)
    {
	cs1 = cs_create_localized (msg_code);
	XtVaSetValues (wid1,
		       XmNmessageString,    cs1,
		       NULL);
	XmStringFree (cs1);
    }

    /*  Pop up the dialog. */
    XtManageChild (wid1);
}

/**************************************************
**      do_menu_choice
**************************************************/
static void do_menu_choice (String tag_string)
{
    Boolean		    popdown_slide;
    String		    str1, str2;
    Widget		    wid1;
    XmString		    cs1;

    popdown_slide = False;
    if (strcmp ((char *) tag_string, "NEW") == 0)
    {
	/* New slide show */
	reclist_clear (main_show);
	popdown_slide = True;
	clear_list (retrieve_widget_id ("slide_list",0));
	clear_entry ();
	set_file_name ("Untitled");

	setup_palette_colors (main_palette,default_color_names);
	update_paledit_info (main_palette);
	main_palette->modified_flag = False;
	setup_default_master_list (main_master_list);
	refresh_master_names (main_master_list);
    }
    else if (strcmp ((char *) tag_string, "OPEN") == 0)
    {
	/*  Open slide show. Pop up file dialog to get
	    slide show file name. */
	cs1 = fetch_mrm_cs ("kcs_open_file_title");
	popup_file_dialog (cs1);
	XmStringFree (cs1);
	popdown_slide = True;
    }
    else if (strcmp ((char *) tag_string, "SAVE") == 0)
    {
	 /* Save slide show */
	str1 = XtNewString (current_file_name);
#ifdef VMS
	/*  On VMS, strip the version number from the file name. */
	str2 = strchr (str1, ';');
	if (str2) *str2 = '\0';
#endif
	write_show (str1,main_show,main_palette,main_master_list);
	XtFree (str1);
    }
    else if (strcmp ((char *) tag_string, "SAVE_AS") == 0)
    {
	/*  "Save as" slide show. Pop up file selection
	    dialog to get file name. */
	cs1 = fetch_mrm_cs ("kcs_save_as_file_title");
	popup_file_dialog (cs1);
	XmStringFree (cs1);
    }
    else if (strcmp ((char *) tag_string, "EXIT") == 0)
    {
	/* Exit program */
	exit (0);
    }

    if (popdown_slide)
    {
	/*  If we are discarding the current show, pop down the slide
	    display form if it is up. */
    	wid1 = retrieve_widget_id ("slide_form",0);
	if (wid1) XtUnmanageChild (wid1);
    }
}

/**************************************************
**      draw_slide
**************************************************/
static void draw_slide
    (slide_info_type *slide, Widget w, Drawable drawable)
{
    Boolean		    bool_data;
    Display		    *disp;
    GC			    draw_gc;
    int			    draw_x, draw_y, pixmap_x, pixmap_y;
    master_element_type	    *ep;
    slide_master_type	    *mp;
    String		    str1;
    unsigned int	    draw_width, draw_height, draw_border,
				draw_depth, element_index,
				pixmap_width, pixmap_height,
				pixmap_border, pixmap_depth;
    Window		    draw_root, pixmap_root;

    /*	This routine draws a slide image into a target drawable,
	either the drawing area window or a buffer pixmap. */

    if (debug_flag)
    {
    	printf ("draw_slide: \n");
    }

    disp = XtDisplay (w);
    mp = find_master (main_master_list,slide->master_name,False);

/*
	To allow for any possible slide master, we should clear the
	drawing area before drawing each slide. However, when the
	masters all start with a common background rectangle, this is
	not really necessary, and it may cause an annoying flash on
	slower X servers when drawing directly to the screen. So we
	check an application resource to decide whether to do the
	clear. If we don't have a master for this slide (maybe it was
	deleted), clear the drawing area in any case. */

    if ((app_resources.clear_bkgnd) || (mp == null_master))
    {
	(void) XGetGeometry (disp,
			     drawable,
			     &draw_root,
			     &draw_x,
			     &draw_y,
			     &draw_width,
			     &draw_height,
			     &draw_border,
			     &draw_depth);
	XFillRectangle (disp,
			drawable,
			draw_clear_gc,
			0,		0,
			draw_width + 1,	draw_height + 1);
    }

    for (element_index = 1;
	 element_index <= mp->elem_list->count;
	 element_index++)
    {
	ep = (master_element_type *)
	    reclist_get_item_ptr (mp->elem_list, element_index);
	if (!ep->element_gc)
	{
	    make_element_gc (w,ep);
	    make_element_pixmap (w,ep);
	}
	draw_gc = ep->element_gc;

	/*  The display of a graphic element is conditional if its
	    definition in the master specifies a Boolean data source.
	    Here, we get the value of that Boolean. */
	switch (ep->element_data)
	{
	    case EL_DATA_HEADER_BAR:
		  bool_data = slide->header_bar;
		  break;
	    case EL_DATA_BODY_BULLET:
		  bool_data =
		    slide->body_lines
			[ep->element_index - 1].body_bullet;
		  break;
	    default:
		  bool_data = True;
	}

	/*  Get the text string to draw, if this is a text element. */
	switch (ep->element_data)
	{
	    case EL_DATA_HEADER_TEXT:
		  str1 = slide->header_text;
		  break;
	    case EL_DATA_TITLE_TEXT:
		  str1 = slide->title_text;
		  break;
	    case EL_DATA_BODY_TEXT:
		  str1 =
		    slide->body_lines
			[ep->element_index - 1].body_text;
		  break;
	    default:
		  str1 = "??";
	}
	str1 = (str1) ? str1 : "";

	/*  Draw the element using the appropriate routine for the
	    element type. */
    	switch (ep->element_type)
	{
	    case EL_TYPE_RECT:
		if (bool_data)
		{
		    /*	Width and height increased by one to
			account for XFillRectangle behavior. */
		    XFillRectangle (disp,
				    drawable,
				    draw_gc,
				    ep->element_x,
				    ep->element_y,
				    ep->element_width + 1,
				    ep->element_height + 1);
		}
		break;
	    case EL_TYPE_TEXT:
		XDrawString (disp,
			     drawable,
			     draw_gc,
			     ep->element_x,
			     ep->element_y,
			     (char *) str1,
			     strlen((char *) str1));
		break;
	    case EL_TYPE_TEXT_STAR:
		XDrawImageString (disp,
				  drawable,
				  draw_gc,
				  ep->element_x,
				  ep->element_y,
				  (char *) str1,
				  strlen((char *) str1));
		break;
	    case EL_TYPE_BULLET:
		if (bool_data)
		{
		    XDrawArc (disp,
			      drawable,
			      draw_gc,
			      ep->element_x,
			      ep->element_y,
			      ep->element_width,
			      ep->element_height,
			      0,
			      360 * 64);
		    XFillArc (disp,
			      drawable,
			      draw_gc,
			      ep->element_x,
			      ep->element_y,
			      ep->element_width,
			      ep->element_height,
			      0,
			      360 * 64);
		}
		break;
	    case EL_TYPE_PIXMAP:
		if ((bool_data) &&
		    (ep->element_pixmap != XmUNSPECIFIED_PIXMAP))
		{
		    (void) XGetGeometry (disp,
					 ep->element_pixmap,
					 &pixmap_root,
					 &pixmap_x,
					 &pixmap_y,
					 &pixmap_width,
					 &pixmap_height,
					 &pixmap_border,
					 &pixmap_depth);
		    if (ep->element_width)
			pixmap_width = ep->element_width;
		    if (ep->element_height)
			pixmap_height = ep->element_height;
		    XCopyArea (disp,
			       ep->element_pixmap,
			       drawable,
			       draw_gc,
			       0,		0,
			       pixmap_width,	pixmap_height,
			       ep->element_x,	ep->element_y);
		}
		break;
	}
    }
}

/**************************************************
**      element_clear
**************************************************/
static void element_clear (master_element_type *ep)
{
    Colormap		cmap;

    if (ep)
    {
    	ep->element_type = EL_TYPE_END;
    	ep->element_x = 0;
    	ep->element_y = 0;
    	ep->element_width = 0;
    	ep->element_height = 0;
	XtFree (ep->element_font_name);
    	ep->element_font_name = (String) NULL;
	XtFree (ep->element_pixmap_name);
    	ep->element_pixmap_name = (String) NULL;
    	ep->element_foreground = 0;
    	ep->element_background = 0;
    	ep->element_data = EL_DATA_NONE;
    	ep->element_index = 0;

	if (ep->element_pixmap != XmUNSPECIFIED_PIXMAP)
	{
	    (void) XmDestroyPixmap (XtScreen (ep->element_gc_widget),
				    ep->element_pixmap);
	    ep->element_pixmap = XmUNSPECIFIED_PIXMAP;
	}

	if (ep->element_gc)
	{
	    XtReleaseGC (ep->element_gc_widget,ep->element_gc);
	    ep->element_gc = (GC) NULL;

	    if (ep->element_gc_mask & GCFont)
	    {
		XUnloadFont (XtDisplay (ep->element_gc_widget),
			     ep->element_font_id);
		ep->element_gc_mask &= (~GCFont);
		ep->element_font_id = 0;
	    }
	}
    ep->element_gc_widget = (Widget) NULL;
    }
}

/**************************************************
**      element_copy
**************************************************/
static void element_copy
    (master_element_type *dst,master_element_type *src)
{
    element_clear (dst);

    dst->element_type = src->element_type;
    dst->element_x = src->element_x;
    dst->element_y = src->element_y;
    dst->element_width = src->element_width;
    dst->element_height = src->element_height;
    dst->element_font_name = XtNewString (src->element_font_name);
    dst->element_pixmap_name = XtNewString (src->element_pixmap_name);
    dst->element_foreground = src->element_foreground;
    dst->element_background = src->element_background;
    dst->element_data = src->element_data;
    dst->element_index = src->element_index;
}

/**************************************************
**      element_create
**************************************************/
static master_element_type *element_create (void)
{
    master_element_type	*ret_ptr;

    ret_ptr = XtNew (master_element_type);

    if (ret_ptr)
    {
    	ret_ptr->element_type = EL_TYPE_END;
    	ret_ptr->element_x = 0;
    	ret_ptr->element_y = 0;
    	ret_ptr->element_width = 0;
    	ret_ptr->element_height = 0;
    	ret_ptr->element_font_name = (String) NULL;
    	ret_ptr->element_pixmap_name = (String) NULL;
    	ret_ptr->element_foreground = 0;
    	ret_ptr->element_background = 0;
    	ret_ptr->element_data = EL_DATA_NONE;
    	ret_ptr->element_index = 0;
    	ret_ptr->element_pixmap = XmUNSPECIFIED_PIXMAP;
    	ret_ptr->element_gc = (GC) NULL;
    	ret_ptr->element_gc_widget = (Widget) NULL;
    	ret_ptr->element_font_id = (Font) NULL;
    	ret_ptr->element_gc_mask = 0;
    }

    return (ret_ptr);
}

/**************************************************
**      element_free
**************************************************/
static void element_free (master_element_type *elem)
{
    element_clear (elem);
    XtFree ((char *) elem);
}

/**************************************************
**      fetch_mrm_cs
**************************************************/
static XmString fetch_mrm_cs (String lit_name)
{
    MrmCode		    lit_type;
    XmString		    lit_value;

    /*	Fetch a compound string from an exported UIL declaration. The
	returned string is a copy and should be freed with
	XmStringFree when no longer needed. */
    if (MrmFetchLiteral (s_MrmHierarchy,
			 lit_name,
			 main_display,
    /* For Motif 1.1, "(caddr_t *) &lit_value" may be required. */
			 (XtPointer *) &lit_value,
			 &lit_type)
	 != MrmSUCCESS)
    {
	printf ("Failed to fetch compound string [%s]\n", lit_name);
	exit (0);
    }

    if (lit_type != MrmRtypeCString)
    {
	printf ("Fetched compound string [%s] has wrong type (%d).\n",
		lit_name, lit_type);
	exit (0);
    }

    return (lit_value);
}

/**************************************************
**      fetch_mrm_int
**************************************************/
static int fetch_mrm_int (String lit_name)
{
    MrmCode		    lit_type;
    int			    lit_value;
    int			    *val_ptr;

    /*	Fetch an integer from an exported UIL declaration. */
    if (MrmFetchLiteral (s_MrmHierarchy,
			 lit_name,
			 main_display,
    /* For Motif 1.1, "(caddr_t *) &val_ptr" may be required. */
			 (XtPointer *) &val_ptr,
			 &lit_type)
	 != MrmSUCCESS)
    {
	printf ("Failed to fetch integer [%s]\n", lit_name);
	exit (0);
    }

    if (lit_type != MrmRtypeInteger)
    {
	printf ("Fetched integer [%s] has wrong type (%d).\n",
		lit_name, lit_type);
	exit (0);
    }

    /*	We received a pointer to the integer value; get the value and
	free the storage allocated by MRM. */
    lit_value = *val_ptr;
    XtFree ((char *) val_ptr);

    return (lit_value);
}

/**************************************************
**      fetch_mrm_string
**************************************************/
static String fetch_mrm_string (String lit_name)
{
    MrmCode		    lit_type;
    String		    lit_value;

    /*	Fetch a string from an exported UIL declaration. The returned
	string is a copy and should be freed with XtFree when no
	longer needed. */
    if (MrmFetchLiteral (s_MrmHierarchy,
			 lit_name,
			 main_display,
    /* For Motif 1.1, "(caddr_t *) &lit_value" may be required. */
			 (XtPointer *) &lit_value,
			 &lit_type)
	 != MrmSUCCESS)
    {
	printf ("Failed to fetch string [%s].\n", lit_name);
	exit (0);
    }

    if (lit_type != MrmRtypeChar8)
    {
	printf ("Fetched string [%s] has wrong type (%d).\n",
		lit_name,
		lit_type);
	exit (0);
    }

    return (lit_value);
}

/**************************************************
**      find_master
**************************************************/
static slide_master_type *find_master (reclist_type *mlp,
				       String mastername,
				       Boolean create_flag)
{
    slide_master_type	*mp;
    unsigned int	indx;

    /*	Look up a slide master by name from a specified slide master
	list. Optionally, create a new slide master with the specified
	name. */

    indx = find_master_index (mlp, mastername);

    if ((!indx) && (create_flag))
    {
	/*  Couldn't find the master, but the caller allows
	    us to create one. */
	mp = master_create ();
	mp->name = XtNewString (mastername);
	reclist_insert_item (mlp, 0, (void *) mp);
	master_free (mp);
	indx = mlp->count;
    }

    /*	If we now have the master, return a pointer to it.
	Otherwise, return a pointer to the null master. */
    if (indx)
    {
	mp = (slide_master_type *)
	    reclist_get_item_ptr (mlp,indx);
    }
    else
    {
	mp = null_master;
    }

    return (mp);
}

/**************************************************
**      find_master_index
**************************************************/
static unsigned int find_master_index (reclist_type *mlp,
				       String mastername)
{
    slide_master_type	*mp;
    unsigned int	indx, found_indx;

    if (!mastername) mastername = "";

    found_indx = 0;
    for (indx = 1;  (!found_indx) && (indx <= mlp->count);  indx++)
    {
    	mp = (slide_master_type *)
	    reclist_get_item_ptr (mlp, indx);
	if (strcmp ((char *) mastername, (char *) mp->name) == 0)
	{
	    found_indx = indx;
	}
    }

    return (found_indx);
}

/**************************************************
**      format_bit_mask
**************************************************/
static String format_bit_mask (unsigned long int mval,
			       int nbits,
			       char c_0,
			       char c_1)
{
    char	*cp;
    String	str1;

    str1 = (String) XtMalloc ((Cardinal) nbits + 1);
    cp = (char *) str1 + nbits;
    *cp-- = '\0';
    while (cp >= str1)
    {
	*cp-- = (mval & 1) ? c_1 : c_0;
	mval >>= 1;
    }
    return (str1);
}

/**************************************************
**      get_element_from_entry
**************************************************/
static void get_element_from_entry (master_element_type *elem)
{
    String		str1;
    Widget		choice_wid;

    str1 = get_option_menu_value
	    (retrieve_widget_id ("mentry_type",0));
    elem->element_type = lookup_el_type (str1);
    XtFree (str1);

    str1 = (String)
	XmTextGetString (retrieve_widget_id ("mentry_xpos",0));
    elem->element_x = atoi (str1);
    XtFree (str1);

    str1 = (String)
	XmTextGetString (retrieve_widget_id ("mentry_ypos",0));
    elem->element_y = atoi (str1);
    XtFree (str1);

    str1 = (String)
	XmTextGetString (retrieve_widget_id ("mentry_width",0));
    elem->element_width = atoi (str1);
    XtFree (str1);

    str1 = (String)
	XmTextGetString (retrieve_widget_id ("mentry_height",0));
    elem->element_height = atoi (str1);
    XtFree (str1);

    XtFree (elem->element_font_name);
    elem->element_font_name = (String)
	XmTextGetString (retrieve_widget_id ("mentry_font",0));

    XtFree (elem->element_pixmap_name);
    elem->element_pixmap_name = (String)
	XmTextGetString (retrieve_widget_id ("mentry_pixmap",0));

    elem->element_foreground = get_option_menu_index
			  (retrieve_widget_id ("mentry_fgcolor",0));
    elem->element_background = get_option_menu_index
			  (retrieve_widget_id ("mentry_bgcolor",0));


    /*	Get the last selected toggle button in the data type
	box. Using its name, set the element data specifier.
	This is awkward because we are using a radio box where
	an option menu would be better; the radio box is used
	for demonstration purposes here. */
    XtVaGetValues (retrieve_widget_id ("mentry_data_type_box",0),
		   XmNmenuHistory,	&choice_wid,
		   NULL);

    /*	Note that the string returned by XtName is not a copy
	and should NOT be deallocated. */
    str1 = XtName (choice_wid);
    if (strcmp ((char *) str1, "elem_data_header_text") == 0)
    {
	elem->element_data = EL_DATA_HEADER_TEXT;
    }
    else if (strcmp ((char *) str1, "elem_data_header_bar") == 0)
    {
	elem->element_data = EL_DATA_HEADER_BAR;
    }
    else if (strcmp ((char *) str1, "elem_data_title_text") == 0)
    {
	elem->element_data = EL_DATA_TITLE_TEXT;
    }
    else if (strcmp ((char *) str1, "elem_data_body_text") == 0)
    {
	elem->element_data = EL_DATA_BODY_TEXT;
    }
    else if (strcmp ((char *) str1, "elem_data_body_bullet") == 0)
    {
	elem->element_data = EL_DATA_BODY_BULLET;
    }
    else
    {
	elem->element_data = EL_DATA_NONE;
    }

    XmScaleGetValue (retrieve_widget_id ("mentry_data_index",0), 
		     &elem->element_index);

    elem->element_gc = (GC) NULL;
    elem->element_gc_widget = (Widget) NULL;

}

/**************************************************
**      get_font_from_name
**************************************************/
static Boolean get_font_from_name (Widget w,
				   String fontname,
				   Font *fontp)
{
    Boolean	ok;
    char	**font_name_list;
    int		font_name_count;

    ok = False;
    if (fontname && fontname[0])
    {
	font_name_list =
	    XListFonts (XtDisplay (w),fontname,1,&font_name_count);

	if (font_name_count)
	{
	    XFreeFontNames (font_name_list);
	    *fontp = XLoadFont (XtDisplay (w),fontname);
	    ok = True;
	}
    }

    return ok;
}

/**************************************************
**      get_option_menu_index
**************************************************/
static Cardinal get_option_menu_index (Widget widopt)
{
    Cardinal		wid_child_count, indx, ret_index;
    Widget		wid_hist, wid_menu;
    WidgetList		wid_children;

    /*	Get widget ID of option menu submenu, and
	of currently selected item. */
    XtVaGetValues (widopt,
		   XmNsubMenuId,	&wid_menu,
		   XmNmenuHistory,	&wid_hist,
		   NULL);

    /* Get list of item widgets in option menu submenu. */
    XtVaGetValues (wid_menu,
		   XmNchildren,	    &wid_children,
		   XmNnumChildren,  &wid_child_count,
		   NULL);

    ret_index = 0;
    for (indx = 1;  indx <= wid_child_count;  indx++)
    {
    	if (wid_hist == wid_children [indx - 1])
	{
	    ret_index = indx;
	    break;
	}
    }

    return (ret_index);
}

/**************************************************
**      get_option_menu_value
**************************************************/
static String get_option_menu_value (Widget widopt)
{
    String		str1;
    Widget		wid_hist;
    XmString		cs1;

    /*	Get widget ID of the currently selected item. */
    XtVaGetValues (widopt,
		   XmNmenuHistory,	&wid_hist,
		   NULL);

    if (wid_hist)
    {
	/*  Get the compound string label from the selected item,
	    and convert it to a string. */
	XtVaGetValues (wid_hist,
		       XmNlabelString,	&cs1,
		       NULL);
	str1 = cs_first_segment (cs1);
	XmStringFree (cs1);
    }
    else
    {
	str1 = (String) NULL;
    }

    return (str1);
}

/**************************************************
**      get_pixel_from_color_num
**************************************************/
static Pixel get_pixel_from_color_num (palette_type *palp,
				       Cardinal cnum)
{
    /*	Return the pixel value corresponding to a
	numbered color in the palette.
	If the color number is unspecified, or out of
	range, substitute the first color in the
	palette. */
    if ((cnum == 0) || (cnum > palp->size)) cnum = 1;

    return (palp->entries[cnum -1].pixel);
}

/**************************************************
**      get_pixel_from_name
**************************************************/
static Boolean get_pixel_from_name (Widget w,
				    String colorname,
				    Pixel *pixelp)
{
    Colormap	cmap;
    Status	status_return;
    XColor	cell_def;

    cmap = DefaultColormapOfScreen (XtScreen(w));
    status_return = 0;
    if (colorname && colorname[0])
    {
	status_return =
	    XParseColor (XtDisplay (w), cmap, colorname, &cell_def);
	if (status_return)
	{
	    status_return =
		XAllocColor (XtDisplay (w), cmap, &cell_def);
	    if (status_return) *pixelp = cell_def.pixel;
	}
    }

    return (status_return != 0);
}

/**************************************************
**      get_slide_from_entry
**************************************************/
static void get_slide_from_entry (slide_info_type *slide)
{
    slide_body_type	*bp1;
    unsigned int	indx;
    Widget		widtmp;
    XmString		cs1;

    XtFree (slide->master_name);
    slide->master_name =
	get_option_menu_value
	    (retrieve_widget_id ("entry_master",0));

    XtFree (slide->header_text);
    slide->header_text = (String) XmTextGetString
			    (retrieve_widget_id ("entry_header",0));

    slide->header_bar =
	XmToggleButtonGetState (retrieve_widget_id ("entry_bar",0));

    XtFree (slide->title_text);
    slide->title_text = (String) XmTextGetString
	(retrieve_widget_id ("entry_title",0));

    bp1 = slide->body_lines;
    for (indx = 1;  indx <= SLIDE_BODY_MAX;  bp1++,indx++)
    {
	bp1->body_bullet = XmToggleButtonGetState
		    (retrieve_widget_id ("entry_bullet", indx));

	XtFree (bp1->body_text);
	bp1->body_text = (String) XmTextGetString
	    (retrieve_widget_id ("entry_body", indx));
    }
}

/**************************************************
**      get_widget_shell
**************************************************/
static Widget get_widget_shell (Widget w)
{
    Widget		wid1;

    /*	Ascend the widget tree until we find
	the shell ancestor of the specified widget. */
    wid1 = w;
    while (!XtIsShell (wid1))
    {
    	wid1 = XtParent (wid1);
    }

    return (wid1);
}

/**************************************************
**      handle_popup_menu
**************************************************/
static void handle_popup_menu
    (Widget	w,
     XtPointer	client_data,
     XEvent	*event,
     Boolean	*continue_to_dispatch)
{
    XButtonPressedEvent	*button_event;
    Widget		wid_popup;
    String		menu_name;

    button_event = (XButtonPressedEvent *) event;
    wid_popup = (Widget) client_data;
    menu_name = XtName (wid_popup);

    /*	Only pop up menu on mouse button 3. */
    if (button_event->button == Button3)
    {
	if (strcmp ((char *) menu_name, "pu_control_menu") == 0)
	{
	    /*  This is the slide editing control pop-up menu;
		enable/disable the menu items according to the
		current processing state. */
	    set_edit_state (main_show->select_index,
			    main_show->count,
			    "pu_control_");
	}
	else if (strcmp ((char *) menu_name, "pu_mctrl_menu") == 0)
	{
	    /*  This is the master editing control pop-up menu;
		enable/disable the menu items according to the
		current processing state. */
	    set_edit_state (edit_master->elem_list->select_index,
			    edit_master->elem_list->count,
			    "pu_mctrl_");
	}

	/*  Set the pop-up menu's position and manage
	    it to make it appear. */
    	XmMenuPosition (wid_popup, button_event);
	XtManageChild (wid_popup);
    }
}

/**************************************************
**      install_image
**************************************************/
static void install_image (String iname,
			   Screen *scrn,
			   int iformat,
			   unsigned int iwidth,
			   unsigned int iheight,
			   unsigned char *idata,
			   palette_type *pal_ptr)
{
    Cardinal		color_num;
    int			ix, iy, ibitmap_pad;
    Pixel		ipix;
    unsigned char	*datap;
    unsigned int	idepth;
    XImage		*image1;

    /*	Quantum for allocating bits to an image row; the value 8
	means each row in the image begins on an even byte. */
    ibitmap_pad = 8;
    idepth = (iformat == XYBitmap) ? 1 : XDefaultDepthOfScreen (scrn);

    image1 = XCreateImage (XDisplayOfScreen (scrn),
			   XDefaultVisualOfScreen (scrn),
			   idepth,
			   iformat,
			   0,
			   (char *) NULL,
			   iwidth,
			   iheight,
			   ibitmap_pad,
			   0);

    if (image1)
    {
    	image1->data = (char *) XtCalloc (image1->height,
					  image1->bytes_per_line);
	datap = idata;
	for (iy = 0;  iy < image1->height;  iy++)
	{
	    for (ix = 0;  ix < image1->width;  ix++)
	    {
		color_num = (Cardinal) *datap++;
		if ((color_num < 1) || (color_num > pal_ptr->size))
		{
		    color_num = 1;
		}
		ipix = pal_ptr->entries[color_num - 1].pixel;
		XPutPixel (image1, ix, iy, ipix);
	    }
	}
	if (!XmInstallImage (image1, (char *) iname))
	{
	    printf ("Failed to install image [%s]\n", iname);
	}
    }
    else
    {
    	printf ("Failed to create image [%s]\n", iname);
    }
}

/**************************************************
**      load_list_from_master
**************************************************/
static void load_list_from_master (slide_master_type *master)
{
    master_element_type	*ep;
    unsigned int	indx;
    Widget		w;
    XmString		cs1;

    w = retrieve_widget_id ("element_list",0);
    clear_list(w);

    for (indx = 1;  indx <= master->elem_list->count;  indx++)
    {
	ep = (master_element_type *)
	    reclist_get_item_ptr (master->elem_list, indx);
	cs1 = cs_create_localized
	    (element_type_names [ep->element_type]);
	XmListAddItemUnselected (w,cs1,0);
	XmStringFree (cs1);
    }
}

/**************************************************
**      load_list_from_show
**************************************************/
static void load_list_from_show (reclist_type *show)
{
    slide_info_type	*sp;
    unsigned int	indx;
    Widget		w;
    XmString		cs1;

    w = retrieve_widget_id ("slide_list",0);
    clear_list(w);

    for (indx = 1;  indx <= show->count;  indx++)
    {
	sp = (slide_info_type *)
	    reclist_get_item_ptr (show, indx);
	put_item_to_list (w,0,sp->title_text);
    }
}

/**************************************************
**      lookup_el_type
**************************************************/
static unsigned int lookup_el_type (String el_str)
{
    unsigned int    indx, found_indx;

    for (indx = 1, found_indx = 0;
	 ((!found_indx) && (indx <= EL_TYPE_MAX));
	 indx++)
    {
    	if (strcmp ((char *) el_str,
		    (char *) element_type_names [indx]) == 0)
	{
	    found_indx = indx;
	}
    }

    return (found_indx);
}

/**************************************************
**      make_element_gc
**************************************************/
static void make_element_gc (Widget		    w,
			     master_element_type    *ep)

{
    XGCValues		gc_values;

    ep->element_gc_mask = 0;
    if (get_font_from_name (w,
			    ep->element_font_name,
			    &gc_values.font))
    {
	ep->element_font_id = gc_values.font;
	ep->element_gc_mask |= GCFont;
    }

    /*	Put the pixel values into the GC data. */
    gc_values.foreground =
	get_pixel_from_color_num (main_palette,
				  ep->element_foreground);
    gc_values.background =
	get_pixel_from_color_num (main_palette,
				  ep->element_background);
    ep->element_gc_mask |= (GCForeground | GCBackground);

    ep->element_gc_widget = w;
    ep->element_gc = XtGetGC (w,
			      ep->element_gc_mask,
			      &gc_values);
}

/**************************************************
**      make_element_pixmap
**************************************************/
static void make_element_pixmap (Widget w,
				 master_element_type *ep)

{
    /*	If a pixmap name was specified, get the corresponding pixmap
	and save it in the element data structure. */
    if ((ep->element_pixmap_name) &&
	(ep->element_pixmap_name [0]))
    {
	ep->element_pixmap =
	    XmGetPixmap (XtScreen (w),
			 (char *) ep->element_pixmap_name,
			 get_pixel_from_color_num
			    (main_palette,ep->element_foreground),
			 get_pixel_from_color_num
			    (main_palette,ep->element_background));
	if (ep->element_pixmap == XmUNSPECIFIED_PIXMAP)
	{
	    printf ("Failed to get pixmap [%s]\n",
		    ep->element_pixmap_name);
	}
    }
    else
    {
	ep->element_pixmap = XmUNSPECIFIED_PIXMAP;
    }
}

/**************************************************
**      master_create
**************************************************/
static slide_master_type *master_create (void)
{
    slide_master_type	*ret_ptr;

    ret_ptr = XtNew (slide_master_type);

    if (ret_ptr)
    {
    	ret_ptr->name = (String) NULL;
	ret_ptr->elem_list = reclist_create
			((reclist_create_fcn) element_create,
			 (reclist_copy_fcn) element_copy);
    }

    return (ret_ptr);
}

/**************************************************
**      master_copy
**************************************************/
static void master_copy (slide_master_type *dst,
			 slide_master_type *src)
{
    unsigned int	indx;
    void		*ep;

    XtFree (dst->name);
    dst->name = XtNewString (src->name);

    reclist_clear (dst->elem_list);

    for (indx = 1;  indx <= src->elem_list->count;  indx++)
    {
	ep = reclist_get_item_ptr (src->elem_list, indx);
	reclist_insert_item (dst->elem_list, 0, ep);
    }
}

/**************************************************
**      master_free
**************************************************/
static void master_free (slide_master_type *master)
{
    master_element_type	    *ep;
    unsigned int	    indx;

    XtFree (master->name);
    master->name = (String) NULL;

    for (indx = 1; indx <= master->elem_list->max_count; indx++)
    {
	ep = (master_element_type *)
	    reclist_get_item_ptr (master->elem_list, indx);
	if (ep)
	{
	    element_free (ep);
	}
    }
    XtFree ((char *) master->elem_list);
    XtFree ((char *) master);
}

/**************************************************
**      mrm_register_name
**************************************************/
static void mrm_register_name (String regname, caddr_t regval)
{
    Cardinal		    mrm_status;
    MrmRegisterArg	    reglist [1];

    /*	Register a name and value for use by the Motif Resource
	Manager. This facility may be used to specify the entry point
	of a callback routine, or to set a value for a UIL
	identifier. */

    if (debug_flag)
    {
	printf ("mrm_register_name: name    [%s]\n",regname);
	printf ("                   value   [%d]\n",regval);
    }

    reglist[0].name = regname;
    reglist[0].value = regval;

    mrm_status = MrmRegisterNames (reglist,1);
    if (mrm_status != MrmSUCCESS)
    {
    	printf ("Unable to register name [%s]\n",regname);
	exit (0);
    }
}

/**************************************************
**      popup_dialog_box
**************************************************/
static Boolean popup_dialog_box (String wid_name,
				 int wid_index,
				 Boolean manage_flag)
{
    Boolean		new_fetch, ok;
    MrmType		widget_class;
    Widget		wid1;

    new_fetch = False;

    /*	See if the dialog box has already been created. */
    wid1 = retrieve_widget_id (wid_name, wid_index);
    ok = (wid1 != (Widget) NULL);

    /*	If not, try to fetch it. */
    if (!ok)
    {
	if (MrmFetchWidget
	    (s_MrmHierarchy,
	     wid_name,
	     wid_top_level,
	     &wid1,
	     &widget_class)
	    == MrmSUCCESS)
	{
	    ok = True;
	    new_fetch = True;
	    save_widget_info (wid_name,wid_index,wid1);
	}
    }

    /*	Manage the dialog box if that is what the caller wants. */
    if ((ok) && (manage_flag))
    {
	XtManageChild (wid1);
    }

    /*	If the fetch failed, terminate with error. */
    if (!ok)
    {
	printf ("popup_dialog_box: failed to fetch [%s]\n", wid_name);
	exit (0);
    }

    /*	Return value is True if we just fetched the dialog, False if
	it had previously been created. */
    return (new_fetch);
}

/**************************************************
**      popup_file_dialog
**************************************************/
static void popup_file_dialog (XmString title_string)
{
    XmString		    cs1;
    Widget		    wid_file_sel;

    (void) popup_dialog_box ("file_selection_box", 0, False);
    wid_file_sel = retrieve_widget_id ("file_selection_box",0);

    /* Get directory mask so we can force search. */
    XtVaGetValues (wid_file_sel,
		   XmNdirMask,	&cs1,
		   NULL);

    XmFileSelectionDoSearch (wid_file_sel,cs1);
    XmStringFree (cs1);

    XtVaSetValues (wid_file_sel,
		   XmNdialogTitle,	    title_string,
		   NULL);

    XtManageChild (wid_file_sel);
}

/**************************************************
**      popup_msg_dialog
**************************************************/
static void popup_msg_dialog (String dialog_name, String body_text)
{
    /*	The maximum length of a UIL name is 31 characters. */
    char		ch1 [32];
    String		body_font_tag;
    Widget		wid1;
    XmString		cs_prefix, cs_suffix, cs_body, cs1, cs2;

    /*	This routine pops up a message dialog and assembles the
	message string. */

    (void) popup_dialog_box (dialog_name, 0, False);
    wid1 = retrieve_widget_id (dialog_name,0);

    /*  Set the message string. It consists of a prefix and suffix,
	fetched from UIL, with the body text in the middle. The body
	text is converted to a compound string, specifying a font tag
	also loaded from UIL. The UIL names are all based on the name
	of the message dialog. */

    sprintf (ch1, "kcs_%s_prefix", dialog_name);
    cs_prefix = fetch_mrm_cs ((String) ch1);
    sprintf (ch1, "kcs_%s_suffix", dialog_name);
    cs_suffix = fetch_mrm_cs ((String) ch1);
    sprintf (ch1, "kch_%s_body", dialog_name);
    body_font_tag = fetch_mrm_string ((String) ch1);

    /*  Convert the body text. */
    cs_body = XmStringCreate ((char *) body_text,
			      (char *) body_font_tag);

    /*  Form prefix + body + suffix. */
    cs1 = XmStringConcat (cs_prefix,cs_body);
    cs2 = XmStringConcat (cs1, cs_suffix);

    /*  Set the message string value. */
    XtVaSetValues (wid1,
		   XmNmessageString,    cs2,
		   NULL);

    /*  Free the storage allocated for the strings. */
    XmStringFree (cs_prefix);
    XmStringFree (cs_suffix);
    XmStringFree (cs_body);
    XmStringFree (cs1);
    XmStringFree (cs2);
    XtFree (body_font_tag);

    /*  Pop up the dialog. */
    XtManageChild (wid1);
}

/**************************************************
**      print_show
**************************************************/
static void print_show (String filnam,
			reclist_type *show,
			reclist_type *master_list)
{
    struct print_show_info  *wpd_ptr;

    /*	Allocate and initialize data structure for
	work procedure. */
    wpd_ptr = XtNew (struct print_show_info);

    wpd_ptr->filename = XtNewString (filnam);
    wpd_ptr->show_ptr = show;
    wpd_ptr->master_list_ptr = master_list;
    wpd_ptr->filep = (FILE *) NULL;
    wpd_ptr->slide_index = 0;

    /*	Request work procedure. */
    (void) XtAppAddWorkProc
		(app_context,
		 (XtWorkProc) wp_print_show,
		 (XtPointer) wpd_ptr);
}

/**************************************************
**      print_slide
**************************************************/
static void print_slide (FILE *out_file,
			 reclist_type *show,
			 unsigned int sindx,
			 reclist_type *master_list)
{
    slide_body_type	*bp;
    slide_info_type	*sp;
    unsigned int	bindx;

    sp = (slide_info_type *)
	reclist_get_item_ptr (show, sindx);
    fprintf (out_file, "%4u. ",sindx);

    /*	For now, we print in a fixed format, ignoring the
	fact that different masters may affect the appearance.
	The master list is available as a parameter in case
	we want to use its information to format the slide. */

    if (sp->header_text)
	fprintf (out_file, "%s\n",sp->header_text);
    else
	fprintf (out_file, "\n");

    if (sp->header_bar)
	fprintf (out_file,
	    "      ---------------------------------\n");
    else
	fprintf (out_file, "\n");

    if (sp->title_text)
	fprintf (out_file, "       %s\n",sp->title_text);
    else
	fprintf (out_file, "\n");

    bp = sp->body_lines;
    for (bindx = 1;  bindx <= SLIDE_BODY_MAX;  bp++,bindx++)
    {
	fprintf (out_file,"        %c ",
		 (bp->body_bullet) ? '*' : ' ');
	if (bp->body_text)
	    fprintf (out_file,"%s\n",bp->body_text);
	else
	    fprintf (out_file, "\n");
    }

    if ((sindx % 7) == 0) fprintf (out_file, "\f");
    fprintf (out_file, "\n");
}

/**************************************************
**      put_element_to_entry
**************************************************/
static void put_element_to_entry (master_element_type *elem)
{
    char		fmt_buff [20];
    Widget		wid1;

    set_option_menu_value
	(retrieve_widget_id ("mentry_type",0),
	 element_type_names [elem->element_type]);

    sprintf (fmt_buff, "%d", elem->element_x);
    put_text_string (retrieve_widget_id ("mentry_xpos",0),fmt_buff);

    sprintf (fmt_buff, "%d", elem->element_y);
    put_text_string (retrieve_widget_id ("mentry_ypos",0),fmt_buff);

    sprintf (fmt_buff, "%u", elem->element_width);
    put_text_string (retrieve_widget_id ("mentry_width",0),fmt_buff);

    sprintf (fmt_buff, "%u", elem->element_height);
    put_text_string (retrieve_widget_id ("mentry_height",0),fmt_buff);

    put_text_string (retrieve_widget_id ("mentry_font",0),
			  elem->element_font_name);
    put_text_string (retrieve_widget_id ("mentry_pixmap",0),
			  elem->element_pixmap_name);

    set_option_menu_index (retrieve_widget_id ("mentry_fgcolor",0),
			   elem->element_foreground);
    set_option_menu_index (retrieve_widget_id ("mentry_bgcolor",0),
			   elem->element_background);

    /*	Now, set the toggle button corresponding to the
	element data source specification.
	This code is awkward because we are using a radio
	box where an option menu would be better; the
	radio box is used to demonstrate its function. */
    switch (elem->element_data)
    {
    	case EL_DATA_HEADER_TEXT:
	    wid1 = retrieve_widget_id ("elem_data_header_text",0);
	    break;
    	case EL_DATA_HEADER_BAR:
	    wid1 = retrieve_widget_id ("elem_data_header_bar",0);
	    break;
    	case EL_DATA_TITLE_TEXT:
	    wid1 = retrieve_widget_id ("elem_data_title_text",0);
	    break;
    	case EL_DATA_BODY_TEXT:
	    wid1 = retrieve_widget_id ("elem_data_body_text",0);
	    break;
    	case EL_DATA_BODY_BULLET:
	    wid1 = retrieve_widget_id ("elem_data_body_bullet",0);
	    break;
    	default:
	    wid1 = retrieve_widget_id ("elem_data_none",0);
    }
    /*	Set the selected toggle button; the third argument to
	XmToggleButtonSetState is set "True" so the same processing
	will occur as if a user selected it. Specifically, this
	ensures that the radio box updates its record of the
	currently selected toggle button. */
    XmToggleButtonSetState (wid1,True,True);

    XmScaleSetValue (retrieve_widget_id ("mentry_data_index",0), 
		     elem->element_index);
}

/**************************************************
**      put_item_to_list
**************************************************/
static void put_item_to_list (Widget w,unsigned int sindx,String str)
{
    XmString	cs1;

    /*	If null or empty (invisible) string, substitute
	something visible. */
    if ((!str) || (!str[0]))
    {
    	str = "= = = = =";
    }
    cs1 = cs_create_localized (str);
    XmListAddItemUnselected (w,cs1,sindx);
    XmStringFree (cs1);

}

/**************************************************
**      put_items_to_list
**************************************************/
static void put_items_to_list (Widget w,
			       unsigned int sindx,
			       String *str_list,
			       unsigned int str_count)
{
    unsigned int	indx;
    XmString		*cs_array;

    cs_array = (XmString *)
	XtCalloc (str_count,sizeof (XmString));

    for (indx = 0;  indx < str_count;  indx++)
    {
	cs_array [indx] = cs_create_localized (str_list [indx]);
    }

    /*	We should really use XmListAddItemsUnselected here, but it
	is not available in Motif versions prior to 1.2. If the
	caller of this routine does a subsequent explicit selection,
	there should be no problem. */
    XmListAddItems (w,cs_array,str_count, (int) sindx);

    for (indx = 0;  indx < str_count;  indx++)
    {
	if (cs_array [indx]) XmStringFree (cs_array [indx]);
    }

    XtFree ((char *) cs_array);

}

/**************************************************
**      put_slide_to_entry
**************************************************/
static void put_slide_to_entry (slide_info_type *slide)
{
    unsigned int	indx;
    slide_body_type	*bp1;

    set_option_menu_value
	(retrieve_widget_id ("entry_master",0),slide->master_name);

    put_text_string
	(retrieve_widget_id ("entry_header",0),slide->header_text);

    /*  No need to notify on this change
	because it's not in a radio box. */
    XmToggleButtonSetState (retrieve_widget_id ("entry_bar",0),
			    slide->header_bar,
			    False);

    put_text_string
	(retrieve_widget_id ("entry_title",0),slide->title_text);

    bp1 = slide->body_lines;
    for (indx = 1;  indx <= SLIDE_BODY_MAX;  bp1++,indx++)
    {
	/*  No need to notify on this change
	    because it's not in a radio box. */
	XmToggleButtonSetState
	    (retrieve_widget_id ("entry_bullet", indx),
	     bp1->body_bullet,
	     False);

	put_text_string (retrieve_widget_id ("entry_body",indx),
			 bp1->body_text);
    }
}

/**************************************************
**      put_text_string
**************************************************/
static void put_text_string (Widget w,String str)
{
    String	    str_old;
    XmTextPosition  ins_pos_old;

    /*	This routine uses XmText* routines that work
	with both XmText and XmTextField widgets.
	It substitutes an empty string if a NULL
	string is passed and checks to see if the
	new value differs from the old to avoid
	unnecessary repainting. It also sets the
	insertion position to the beginning of the
	string. */

    str_old = (String) XmTextGetString (w);

    if (str)
    {
	if (strcmp ((char *) str, (char *) str_old))
	{
	    XmTextSetString (w,str);
	}
    }
    else
    {
	if (str_old[0])
	{
	    XmTextSetString (w,"");
	}
    }

    ins_pos_old = XmTextGetInsertionPosition(w);
    if (ins_pos_old)
    {
	XmTextSetInsertionPosition (w,0);
    }

    XtFree (str_old);
}

/**************************************************
**      reclist_clear
**************************************************/
static void reclist_clear (reclist_type *rlist)
{
    if (rlist)
    {
    	rlist->count = 0;
    	rlist->select_index = 0;
    	rlist->max_count = 0;
    	rlist->modified_flag = False;
    }
}

/**************************************************
**      reclist_create
**************************************************/
static reclist_type *reclist_create
    (reclist_create_fcn	    create_fcn,
     reclist_copy_fcn	    copy_fcn)
{
    reclist_type	    *ret_ptr;

    ret_ptr = XtNew (reclist_type);

    ret_ptr->count = 0;
    ret_ptr->select_index = 0;
    ret_ptr->max_count = 0;
    ret_ptr->modified_flag = False;
    ret_ptr->recs = (void **) NULL;
    ret_ptr->fcn_create = create_fcn;
    ret_ptr->fcn_copy = copy_fcn;

    return (ret_ptr);
}

/**************************************************
**      reclist_delete_item
**************************************************/
static void reclist_delete_item
    (reclist_type *rlist, unsigned int delindx)
{
    unsigned int	indx;
    void		*rp1, **rpp1;

    if ((delindx > 0) && (delindx <= rlist->count))
    {
	if (delindx == rlist->select_index)
	{
	    rlist->select_index = 0;
	}

	/*  Get pointer to pointer of item being deleted. */
	rpp1 = rlist->recs + delindx - 1;

	/*  Save pointer we are overwriting. */
	rp1 = *rpp1;

	/*  Copy pointers down to close up list;
	    if at end, no copy needed. */
	for (indx = delindx;  indx < rlist->count;  indx++)
	{
	    *rpp1 = *(rpp1 + 1);
	    rpp1++;
	}

	/* Put the saved pointer into the new top slot. */
	*rpp1 = rp1;

	rlist->count--;
	rlist->modified_flag = True;
    }

}

/**************************************************
**      reclist_get_item_ptr
**************************************************/
static void *reclist_get_item_ptr
    (reclist_type *rlist, unsigned int rindx)
{
    void	*ret_ptr;

    if ((rlist) &&
	(rindx > 0) &&
	(rindx <= rlist->count))
    {
	ret_ptr = *(rlist->recs + rindx - 1);
    }
    else
    {
	ret_ptr = (void *) NULL;
    }

    return (ret_ptr);
}

/**************************************************
**      reclist_insert_item
**************************************************/
static void reclist_insert_item (reclist_type *rlist,
				 unsigned int insindx,
				 void *insrec)
{
    unsigned int	indx, new_max, new_size;
    void		*rp1, **rpp1;

    if (rlist->count >= rlist->max_count)
    {
	new_max = (rlist->max_count + RECLIST_ALLOC_DELTA);
	new_size = new_max * sizeof (void *); 
    	if (rlist->max_count == 0)
	{
	    rpp1 = (void **) XtMalloc (new_size);
	}
	else
	{
	    rpp1 = (void **)
		XtRealloc ((char *) rlist->recs, new_size);
	}

	rlist->recs = rpp1;

	/* Bump pointer to beginning of newly allocated area. */
	rpp1 += rlist->max_count;
	while (rlist->max_count < new_max)
	{
	    /* Clear out pointers in newly allocated area. */
	    *(rpp1++) = (void *) NULL;
	    rlist->max_count++;
	}
    }

    rlist->modified_flag = True;
    rlist->count++;
    /* If index is zero, insert after all existing items. */
    insindx = (insindx) ?
	MIN (insindx, rlist->count) : rlist->count;

    /*  Get pointer to pointer of (new) highest record in list. */
    rpp1 = rlist->recs + rlist->count - 1;

    /*  Save pointer we are overwriting. */
    rp1 = *rpp1;

    /*  Copy pointers up to make room for entry;
	if at end, no copy needed. */
    for (indx = rlist->count;  indx > insindx;  indx--)
    {
    	*rpp1 = *(rpp1 - 1);
	rpp1--;
    }

    /* If the saved pointer was NULL, allocate a record. */
    if (!rp1)
    {
	rp1 = (*(rlist->fcn_create)) ();
    }

    /* Put the saved pointer into the newly opened slot. */
    *rpp1 = rp1;

    /* Copy in the data to be inserted. */
    (*(rlist->fcn_copy)) (rp1,insrec);
}

/**************************************************
**      reclist_replace_item
**************************************************/
static void reclist_replace_item (reclist_type *rlist,
				  unsigned int repindx,
				  void *reprec)
{
    void		*rp1;

    rp1 = reclist_get_item_ptr (rlist,repindx);

    if (rp1)
    {
	(*(rlist->fcn_copy)) (rp1,reprec);
	rlist->modified_flag = True;
    }
}

/**************************************************
**      refresh_master_names
**************************************************/
static void refresh_master_names (reclist_type *mlp)
{
    slide_master_type	*mp;
    unsigned int	indx;
    Widget		wid1, wid2;

    wid1 = retrieve_widget_id ("entry_master",0);
    wid2 = retrieve_widget_id ("edit_menu_edit_master",0);

    clear_menu_buttons (wid1);
    clear_menu_buttons (wid2);

    for (indx = 1;  indx <= mlp->count;  indx++)
    {
	mp = (slide_master_type *)
	    reclist_get_item_ptr (mlp, indx);
    	add_menu_button (wid1,mp->name);
    	add_menu_button (wid2,mp->name);
    }

    set_option_menu_value (wid1,"");
}

/**************************************************
**      replace_item_in_list
**************************************************/
static void replace_item_in_list
    (Widget w, unsigned int sindx, String str)
{
    XmString	cs1;

    cs1 = cs_create_localized (str);
    XmListReplaceItemsPos (w,&cs1,1,sindx);
    XmStringFree (cs1);

}

/**************************************************
**      retrieve_widget_id
**************************************************/
static Widget retrieve_widget_id (String widnam, int wididx)
{
    Widget		widid;
    widget_entry_type	*wp;

    widid = (Widget) NULL;
    if (widnam)
    {
	wp = wid_list_head;
	while ((wp) && (!widid))
	{
	    if ((wp->wid_index == wididx) &&
		(strcmp ((char *)wp->wid_name,
			 (char *) widnam)
		 == 0))
	    {
		widid = wp->wid_id;
	    }
	    wp = wp->next;
	}
    }

    return (widid);
}

/**************************************************
**      retrieve_widget_info
**************************************************/
static widget_entry_type *retrieve_widget_info (Widget w)
{
    widget_entry_type	*wp_ret, *wp;

    wp_ret = (widget_entry_type *) NULL;
    wp = wid_list_head;
    while ((wp) && (!wp_ret))
    {
	if (wp->wid_id == w)
	{
	    wp_ret = wp;
	}
	wp = wp->next;
    }

    return (wp_ret);
}

/**************************************************
**      save_widget_info
**************************************************/
static void save_widget_info (String widnam,
			      int wididx,
			      Widget widid)
{
    widget_entry_type	*wp, *wp2;

    wp2 = (widget_entry_type *) NULL;
    wp = wid_list_head;
    while ((wp) && (!wp2))
    {
    	if ((wp->wid_index == wididx) &&
	    (strcmp ((char *)wp->wid_name,
		     (char *) widnam)
	     == 0))
	{
	    wp2 = wp;
	}
	wp = wp->next;
    }

    if (!wp2)
    {
    	wp2 = XtNew (widget_entry_type);
	if (wp2)
	{
	    wp2->wid_name = XtNewString (widnam);
	    wp2->wid_index = wididx;
	    wp2->next = wid_list_head;
	    wid_list_head = wp2;
	}
    }

    if (wp2)
    {
	wp2->wid_id = widid;
    }
}

/**************************************************
**      select_item_in_list
**************************************************/
static void select_item_in_list (Widget w, unsigned int sindx)
{
    int		top_pos;
    int		vis_cnt;

    XmListSelectPos (w,sindx,True);

    XtVaGetValues (w,
		   XmNtopItemPosition,	    &top_pos,
		   XmNvisibleItemCount,	    &vis_cnt,
		   NULL);

    if ((int) sindx < top_pos)
    {
    	XmListSetPos (w,sindx);
    }
    else if ((int) sindx >= top_pos + vis_cnt)
    {
    	XmListSetBottomPos (w,sindx);
    }
}

/**************************************************
**      set_cursor
**************************************************/
static void set_cursor (Widget w, unsigned int cursor_shape)
{
    Cursor	curs;
    Display	*disp;
    Window	win;

    disp = XtDisplay (w);
    win = XtWindow (w);
    if (cursor_shape != DEFAULT_CURSOR)
    {
	curs = XCreateFontCursor (disp,cursor_shape);
	if (curs != (Cursor) None)
	{
	    XDefineCursor (disp,win,curs);
	    XFreeCursor (disp,curs);
	}
    }
    else
    {
    	XUndefineCursor (disp,win);
    }
}

/**************************************************
**      set_disp_slide
**************************************************/
static void set_disp_slide (unsigned int indx)
{
    char		    ch1[30];
    slide_info_type	    *sp1;
    XmString		    cs1;
    Widget		    wid1;
    Window		    win1;

    indx = MIN (indx,main_show->count);
    disp_indx = indx;

    if (indx)
    {
	sp1 = (slide_info_type *)
	    reclist_get_item_ptr (main_show, indx);
	slide_copy (disp_slide,sp1);
    }
    else
    {
	slide_clear (disp_slide);
    }

    sprintf (ch1, "Slide %u", indx);
    cs1 = cs_create_localized ((String) ch1);
    XtVaSetValues (retrieve_widget_id ("slide_number",0),
		   XmNlabelString,	cs1,
		   NULL);
    XmStringFree (cs1);

    update_disp_slide (disp_slide);
}

/**************************************************
**      set_edit_state
**************************************************/
static void set_edit_state (unsigned int sel_indx,
			    unsigned int cnt,
			    char *prefix)
{
    /*	The maximum length of a UIL name is 31 characters. */
    char	widname[32], *suffixp;
    Widget	wid1;

    /*	Set the buttons in an editing control menu to be
	enabled/disabled according to the state of the
	aggregate being edited. The "sel_indx" and
	"cnt" arguments specify the selected item and
	the total number of items in the aggregate.
	The menu button widgets must be named with
	certain fixed suffixes according to the button's
	function; the prefix portion of each menu button
	name must be the same, corresponding to the
	"prefix" argument. */

    strcpy (widname, prefix);
    suffixp = &widname [strlen (prefix)];

    /* "Get previous" choice. */
    strcpy (suffixp, "get_previous");
    wid1 = retrieve_widget_id (widname, 0);
    XtSetSensitive (wid1,(sel_indx > 1));

    /* "Get selected" choice. */
    strcpy (suffixp, "get_selected");
    wid1 = retrieve_widget_id (widname, 0);
    XtSetSensitive (wid1,(sel_indx > 0));

    /* "Get next" choice. */
    strcpy (suffixp, "get_next");
    wid1 = retrieve_widget_id (widname, 0);
    XtSetSensitive (wid1,(sel_indx < cnt));

    /* "Insert before" choice. */
    strcpy (suffixp, "insert_before");
    wid1 = retrieve_widget_id (widname, 0);
    XtSetSensitive (wid1,(sel_indx > 0));

    /* "Replace" choice. */
    strcpy (suffixp, "replace");
    wid1 = retrieve_widget_id (widname, 0);
    XtSetSensitive (wid1,(sel_indx > 0));

    /* "Delete" choice. */
    strcpy (suffixp, "delete");
    wid1 = retrieve_widget_id (widname, 0);
    XtSetSensitive (wid1,(sel_indx > 0));
}

/**************************************************
**      set_file_name
**************************************************/
static void set_file_name (String filnam)
{
    String		str1;

    XtFree (current_file_name);
    current_file_name = XtNewString((filnam) ? filnam : "");

    str1 = (String)
	XtMalloc (strlen (app_name) +
		  strlen (current_file_name) + 3 + 1);

    strcpy ((char *) str1, (char *) app_name);
    strcat ((char *) str1, " - ");
    strcat ((char *) str1, (char *) current_file_name);

    XtVaSetValues (wid_top_level,
		   XmNtitle,	    str1,
		   XmNiconName,	    current_file_name,
		   NULL);

    XtFree (str1);
}

/**************************************************
**      set_icon_bitmap
**************************************************/
#ifndef APP_ICON_UIL
static void set_icon_bitmap (Widget w,
			     int iconwidth,
			     int iconheight,
			     char *icondata)
{
    Pixmap		    icon_pixmap;

    /*	Create the pixmap from the binary data. */
    icon_pixmap = XCreateBitmapFromData
	(XtDisplay (w),
	 XtWindow (w),
	 icondata,
	 iconwidth,
	 iconheight);

    if (icon_pixmap)
    {
	/*  Set the bitmap as the application icon.
	    The window manager (mwm) actually does the work,
	    based on the data we supply. */
	XtVaSetValues (w,
		       XmNiconPixmap,	icon_pixmap,
		       NULL);
    }
}
#endif

/**************************************************
**      set_icon_uil
**************************************************/
#ifdef APP_ICON_UIL
static void set_icon_uil (Widget w, String iconname)
{
    Boolean		    ok;
    Dimension		    icon_width, icon_height;
    Pixmap		    icon_pixmap;

    /*	Fetch the UIL bitmap. */
    ok = (MrmFetchBitmapLiteral
	    (s_MrmHierarchy,
	     iconname,
	     XtScreen (w),
	     XtDisplay (w),
	     &icon_pixmap,
	     &icon_width,
	     &icon_height)
	 == MrmSUCCESS);

    if (ok)
    {
	/*  Set the bitmap as the application icon.
	    The window manager (mwm) actually does the work,
	    based on the data we supply. */
	XtVaSetValues (w,
		       XmNiconPixmap,	icon_pixmap,
		       NULL);
    }
}
#endif

/**************************************************
**      set_master_edit_name
**************************************************/
static void set_master_edit_name (String master_name)
{
    Cardinal		    str_len;
    String		    str1,str2;
    unsigned int	    indx;
    XmString		    cs1;

    /*	Set the new master name in the master data structure. */
    XtFree (edit_master->name);
    edit_master->name = XtNewString (master_name);

    /*	Set the dialog box title. */
    str1 = fetch_mrm_string ("ks_edit_master_title");
    str_len = strlen ((char *) str1) +
	      strlen ((char *) master_name) + 1;
    str2 = (String) XtMalloc (str_len);
    strcpy ((char *) str2, (char *) str1);
    strcat ((char *) str2, (char *) master_name);
    cs1 = cs_create_localized (str2);
    XtVaSetValues (retrieve_widget_id ("master_form",0),
		   XmNdialogTitle,	cs1,
		   NULL);
    XtFree (str1);
    XtFree (str2);
    XmStringFree (cs1);

    /*	Enable the "Delete master" button only if the
	master being edited already exists and is not
	the only master. */
    indx = find_master_index (main_master_list,edit_master->name);
    XtSetSensitive (retrieve_widget_id ("master_delete",0),
		    (indx > 0) &&
		    (main_master_list->count > 1));
}

/**************************************************
**      set_option_menu_index
**************************************************/
static void set_option_menu_index (Widget widopt, Cardinal indx)
{
    Cardinal		wid_child_count;
    Widget		wid_menu, wid_hist, *wid_child_ptr;
    WidgetList		wid_children;

    /* Get widget ID of option menu submenu. */
    XtVaGetValues (widopt,
		   XmNsubMenuId,	&wid_menu,
		   NULL);

    /* Get list of item widgets in option menu submenu. */
    XtVaGetValues (wid_menu,
		   XmNchildren,	    &wid_children,
		   XmNnumChildren,  &wid_child_count,
		   NULL);

    if ((indx > 0) && (indx <= wid_child_count))
    {
    	wid_hist = wid_children [indx - 1];
    }
    else
    {
	wid_hist = (Widget) NULL;
    }

    /* Set the menu history resource of the option menu. */
    if ((!wid_hist) && (wid_child_count)) wid_hist = *wid_children;
    XtVaSetValues (widopt,
		   XmNmenuHistory,	wid_hist,
		   NULL);

}

/**************************************************
**      set_option_menu_value
**************************************************/
static void set_option_menu_value (Widget widopt,String strval)
{
    Cardinal		indx, wid_child_count;
    String		str1;
    Widget		wid_menu, wid_hist, *wid_child_ptr;
    WidgetList		wid_children;
    XmString		cs1, cs2;

    if (!strval) strval = "";

    /* Get widget ID of option menu submenu. */
    XtVaGetValues (widopt,
		   XmNsubMenuId,	&wid_menu,
		   NULL);

    /* Get list of item widgets in option menu submenu. */
    XtVaGetValues (wid_menu,
		   XmNchildren,	    &wid_children,
		   XmNnumChildren,  &wid_child_count,
		   NULL);

    /*	Make compound string version of the string
	we are looking for. */
    cs2 = cs_create_localized (strval);

    /* Search through items, looking for a matching label string.
	If found, save the widget ID to set the option
	menu history. */
    indx = 1;
    wid_child_ptr = wid_children;
    wid_hist = (Widget) NULL;
    while ((!wid_hist) && (indx <= wid_child_count))
    {
	XtVaGetValues (*wid_child_ptr,
		       XmNlabelString,	    &cs1,
		       NULL);

	if (XmStringCompare (cs1,cs2))
	{
	    wid_hist = *wid_child_ptr;
	}
	XmStringFree (cs1);
	wid_child_ptr++;
	indx++;
    }
    XmStringFree (cs2);

    /* Set the menu history resource of the option menu. */
    if ((!wid_hist) && (wid_child_count)) wid_hist = *wid_children;
    XtVaSetValues (widopt,
		   XmNmenuHistory,	wid_hist,
		   NULL);

}

/**************************************************
**      setup_default_master
**************************************************/
static void setup_default_master (reclist_type *mlp,
				  String mastername,
				  master_element_type *srclist)
{
    slide_master_type	    *mp;

    /*	Create a new slide master. */
    mp = master_create ();

    if (mp)
    {
	/*  Set the new slide master name. */
	mp->name = XtNewString (mastername);

	/*  Load a list of elements into the
	    newly created slide master. */
	while (srclist->element_type != EL_TYPE_END)
	{
	    reclist_insert_item (mp->elem_list,
				 0,
				 (void *) srclist++);
	}

	mp->elem_list->modified_flag = False;

	/*  Insert (copy) the new slide master into the
	    caller's slide master list. */
    	reclist_insert_item (mlp, 0, (void *) mp);

	/*  Free the temporary slide master storage. */
	master_free (mp);
    }

}

/**************************************************
**	setup_default_master_list
**************************************************/
static void setup_default_master_list (reclist_type *mlp)
{
    /*	Load a slide master list with a set of default slide masters.
	First, clear out any existing masters in the master list. */
    reclist_clear (mlp);

    /*	Create default slide masters and load them into the caller's
	slide master list. */
    setup_default_master (mlp, "Body 1", default_body_master);
    setup_default_master (mlp, "Title", default_title_master);
    mlp->modified_flag = False;
}

/**************************************************
**	setup_palette
**************************************************/
static palette_type *setup_palette (Widget w,
				    int psize)
{
    Cardinal		ii;
    palette_entry_type	*pep1;
    palette_type	*pp1;
    Pixel		fg_pixel, bg_pixel;

    /*	Note that the argument "psize" is declared as an "int";
	we actually want a "Cardinal" value, but then we would
	have to set up a type converter for "String" to "Cardinal".
	So we validate the argument as an integer and then convert
	it to "Cardinal" for use. */
    if (psize < PALETTE_SIZE_MIN) psize = PALETTE_SIZE_DEFAULT;

    pp1 = XtNew (palette_type);
    pp1->size = (Cardinal) psize;
    pp1->modified_flag = False;
    pp1->screen = XtScreen (w);
    pp1->entries = (palette_entry_type *)
	XtCalloc (pp1->size, sizeof (palette_entry_type));

    /*	Get the default foreground and background colors from the
	specified widget. Note that only subclasses of XmPrimitive
	and XmManager have an XmNforeground resource, so the
	caller must provide a widget of the proper type.
	This widget is also used to identify the screen where the
	colors should be allocated, and the created palette will
	only be valid on that screen. */
    fg_pixel = bg_pixel = (Pixel) 0;
    XtVaGetValues (w,
		   XmNforeground,   &fg_pixel,
		   XmNbackground,   &bg_pixel,
		   NULL);

    /*	Set up the color palette entries. */
    pep1 = pp1->entries;
    for (ii = 0;  ii < pp1->size;  pep1++,ii++)
    {
    	pep1->color_name = (String) NULL;

	/*  Try to allocate an exclusive read/write color cell
	    for this color. If we fail, use the background color
	    pixel instead, but remember that we should not change
	    its value (allocated_pixel == False). */
    	pep1->pixel = bg_pixel;
    	pep1->allocated_pixel = alloc_color_cell (w,&pep1->pixel);

	/*  If we allocated a color cell successfully, create a
	    sample color pixmap for future use. The pixmap uses the
	    entry color as its foreground, and the default foreground
	    color as its background, to make contrast more likely. */
	if (pep1->allocated_pixel)
	{
	    /*	Try to make a pixmap using our special
		color sample pattern. */
	    pep1->sample_pixmap =
		    XmGetPixmap (pp1->screen,
				 "color_sample",
				 pep1->pixel,
				 fg_pixel);

	    /*	If we failed to get our pixmap, try to do it with the
		fallback pattern. Because the fallback pattern is all
		"background" color, we reverse the foreground and
		background pixel specifications. (We could use our
		pixel value for both foreground and background in this
		case.) */
	    if (pep1->sample_pixmap == XmUNSPECIFIED_PIXMAP)
	    {
		pep1->sample_pixmap =
			XmGetPixmap (pp1->screen,
				     "background",
				     fg_pixel,
				     pep1->pixel);
	    }
	}
	else
	{
	    pep1->sample_pixmap = XmUNSPECIFIED_PIXMAP;
	}
    }
    return (pp1);
}

/**************************************************
**	setup_palette_colors
**************************************************/
static void setup_palette_colors (palette_type *pal_ptr,
				  String *cnames)
{
    Cardinal		ii;
    palette_entry_type	*pep1;
    String		*cnamep;

    /*	Set up the color palette entries, using the supplied
	color names. */
    cnamep = cnames;
    pep1 = pal_ptr->entries;
    for (ii = 0;  ii < pal_ptr->size;  pep1++,ii++)
    {
	/*  Get the next color name and assign it to this palette
	    entry. If we hit the end of the name list, the
	    rest of the entries will have a null name. */
	XtFree (pep1->color_name);
    	pep1->color_name = XtNewString (*cnames);
	if (*cnames) cnames++;

	/*  If we allocated a color cell successfully, store the
	    named color in it. */
	if (pep1->allocated_pixel)
	{
	    (void) store_color_in_cell
		    (pal_ptr->screen, pep1->color_name, pep1->pixel);
	}
    }
}

/**************************************************
**	setup_popup_menu
**************************************************/
static void setup_popup_menu (String widname,
			      Widget parent)
{
    Cardinal		mrm_status;
    MrmType		widget_class;
    Widget		pop_menu;

    /*	Initialize widget variable to NULL, since some (old)
	versions of Motif no-op the fetch otherwise. */
    pop_menu = (Widget) NULL;

    mrm_status = MrmFetchWidget (s_MrmHierarchy,
				 widname,
				 parent,
				 &pop_menu,
				 &widget_class);

    if (mrm_status != MrmSUCCESS)
    {
    	printf ("Can't fetch pop-up menu '%s'.\n", widname);
	exit (0);
    }

    XtAddEventHandler (parent,
		       (EventMask) ButtonPressMask,
		       False,
		       (XtEventHandler) handle_popup_menu,
		       (XtPointer) pop_menu);

}

/**************************************************
**      slide_clear
**************************************************/
static void slide_clear (slide_info_type *slide)
{
    slide_body_type	*bp1;
    unsigned int	indx;

    XtFree (slide->master_name);
    slide->master_name = (String) NULL;

    XtFree (slide->header_text);
    slide->header_text = (String) NULL;

    slide->header_bar = False;

    XtFree (slide->title_text);
    slide->title_text = (String) NULL;

    bp1 = slide->body_lines;
    for (indx = 1;  indx <= SLIDE_BODY_MAX;  bp1++,indx++)
    {
    	bp1->body_bullet = False;
	XtFree (bp1->body_text);
	bp1->body_text = (String) NULL;
    }
}

/**************************************************
**      slide_copy
**************************************************/
static void slide_copy (slide_info_type *dst,slide_info_type *src)
{
    slide_body_type	*bps, *bpd;
    unsigned int	indx;

    slide_clear (dst);

    dst->master_name = XtNewString (src->master_name);
    dst->header_text = XtNewString (src->header_text);
    dst->header_bar = src->header_bar;
    dst->title_text = XtNewString (src->title_text);

    bps = src->body_lines;
    bpd = dst->body_lines;
    for (indx = 1;  indx <= SLIDE_BODY_MAX;  bps++,bpd++,indx++)
    {
    	bpd->body_bullet = bps->body_bullet;
	bpd->body_text = XtNewString (bps->body_text);
    }
}

/**************************************************
**      slide_create
**************************************************/
static slide_info_type *slide_create (void)
{
    slide_body_type	*bp;
    slide_info_type	*ret_ptr;
    unsigned int	indx;

    ret_ptr = XtNew (slide_info_type);

    if (ret_ptr)
    {
    	ret_ptr->master_name = (String) NULL;
    	ret_ptr->header_text = (String) NULL;
    	ret_ptr->header_bar = False;
    	ret_ptr->title_text = (String) NULL;

	bp = ret_ptr->body_lines;
	for (indx = 1;  indx <= SLIDE_BODY_MAX;  indx++)
	{
	    bp->body_bullet = False;
	    bp->body_text = (String) NULL;
	    bp++;
	}
    }

    return (ret_ptr);
}

/**************************************************
**      slide_free
**************************************************/
static void slide_free (slide_info_type *slide)
{
    slide_clear (slide);
    XtFree ((char *) slide);
}

/**************************************************
**      store_color_in_cell
**************************************************/
static Boolean store_color_in_cell (Screen *scrn,
				    String cname,
				    Pixel pix)
{
    Boolean		    ok;
    Colormap		    cmap1;
    Status		    ret_status;
    XColor		    color1;

    cmap1 = XDefaultColormapOfScreen (scrn);

    ok = (cname != (String) NULL);

    if (ok)
    {
	ret_status = XParseColor (XDisplayOfScreen (scrn),
				  cmap1,
				  (char *) cname,
				  &color1);
	ok = (ret_status != 0);
    }

    if (!ok)
    {
	/*  If color name was not found, set to black. */
	color1.red = 0;
	color1.green = 0;
	color1.blue = 0;
	color1.flags = DoRed | DoGreen | DoBlue;
    }

    color1.pixel = pix;
    XStoreColor (XDisplayOfScreen (scrn),
		 cmap1,
		 &color1);

    return (ok);
}

/**************************************************
**      update_disp_slide
**************************************************/
static void update_disp_slide (slide_info_type *slide)
{
    Dimension		    wid_width, wid_height;
    unsigned int	    win_width, win_height;
    Widget		    wid1;
    Window		    win1;

    wid1 = retrieve_widget_id ("slide_area",0);
    win1 = XtWindow (wid1);

    if (app_resources.use_buffer_pixmap)
    {
	/*  Draw the slide into the pixmap. */
	draw_slide (slide, wid1, (Drawable) buffer_pixmap);

	/*  Now we must copy the pixmap to the window if it exists
	    (the drawing area may not yet be realized).
	    We could just call XClearArea, setting the "exposures"
	    argument to True and let the exposure callback take
	    care of the copy. However, this may cause an annoying
	    flash on slow servers when the window is cleared
	    before being redrawn, so we do the copy ourselves here. */
	if (win1)
	{
	    XtVaGetValues (wid1,
			   XmNwidth,    &wid_width,
			   XmNheight,   &wid_height,
			   NULL);
	    win_width = wid_width;
	    win_height = wid_height;
	    XCopyArea (XtDisplay (wid1),
		       buffer_pixmap,
		       win1,
		       XDefaultGCOfScreen (XtScreen (wid1)),
		       0,		0,
		       win_width,	win_height,
		       0,		0);
	}
    }
    else
    {
	/*  Draw the slide directly to the window if the widget
	    is realized (has a window). It may not be realized
	    yet if we are now popping up the slide display dialog. */
	if (win1) draw_slide (slide, wid1, (Drawable) win1);
    }
}

/**************************************************
**      update_paledit_info
**************************************************/
static void update_paledit_info (palette_type *pal_ptr)
{
    Widget		wid1;
    unsigned int	ii;

    /*	Display the color names from the color palette entries
	in the palette edit dialog. */
    for (ii = 1;  ii <= pal_ptr->size;  ii++)
    {
	/*  Set the color index number. */
	wid1 = retrieve_widget_id ("paledit_color_name",ii);
	if (wid1) put_text_string
			(wid1, pal_ptr->entries[ii - 1].color_name);
    }
}

/**************************************************
**      update_show_status
**************************************************/
static void update_show_status (reclist_type *show)
{
    /*	Update some aspects of the user interface to reflect the
	current editing status of the slide show.
	Set the available options in the editing control menu.
	Format the slide list "slide m of n" message.
	Enable/disable the "print" facility. */

    set_edit_state (show->select_index, show->count, "control_");
    update_slide_list_descr (show->select_index, show->count);
    XtSetSensitive (retrieve_widget_id ("file_menu_print",0),
		    (show->count > 0));
    XtSetSensitive (retrieve_widget_id ("main_icon_bar_print",0),
		    (show->count > 0));
}

/**************************************************
**      update_slide_list_descr
**************************************************/
static void update_slide_list_descr (unsigned int sel_indx,
				     unsigned int cnt)
{
    char	fmt_buff [100];
    Pixel	old_bkgnd, new_bkgnd;
    Widget	wid1;

    /*	Update the "slide n of m" description for the slide list. The
	message format should reference (e.g., "%d") the select index
	and count if the select index is non-zero, and only the count
	otherwise.
	These message formats are read at run time from UIL. */
    sprintf (fmt_buff,
	     sel_indx ? ks_slide_list_descr_sel :
			ks_slide_list_descr_nosel,
	     sel_indx ? sel_indx : cnt,
	     cnt);
    wid1 = retrieve_widget_id ("slide_list_descr",0);
    put_text_string (wid1, (String) fmt_buff);

    /*	Set the background color for the message.
	First, get the current value so we don't cause an
	unnecessary redraw. */
    new_bkgnd = sel_indx ? slide_descr_pixel_sel :
			   slide_descr_pixel_nosel;
    XtVaGetValues (wid1,
		   XmNbackground,   &old_bkgnd,
		   NULL);
    if (new_bkgnd != old_bkgnd)
    {
    	XtVaSetValues (wid1,
		       XmNbackground,	new_bkgnd,
		       NULL);
    }
}

/**************************************************
**      wp_print_show
**************************************************/
static Boolean wp_print_show (XtPointer client_data)
{
    Boolean		    done, ok;
    FILE		    *out_file;
    reclist_type	    *show, *master_list;
    struct print_show_info  *dp;

    dp = (struct print_show_info *) client_data;
    show = dp->show_ptr;
    master_list = dp->master_list_ptr;
    out_file = dp->filep;
    done = False;
    ok = True;

    if (!out_file)
    {
	if (debug_flag)
	{
	    printf ("wp_print_show: initializing.\n");
	}

	set_cursor (wid_main_window, XC_watch);

	out_file = fopen (dp->filename, "w");
	dp->filep = out_file;
	ok = (out_file != (FILE *) NULL);
    }
    else if (dp->slide_index < show->count)
    {
	dp->slide_index++;

	if (debug_flag)
	{
	    printf ("wp_print_show: printing slide %d.\n",
		    dp->slide_index);
	}
	print_slide (out_file, show, dp->slide_index, master_list);
    }
    else
    {
	done = True;
    }

    if (!ok) done = True;

    if (done)
    {
	if (debug_flag)
	{
	    printf ("wp_print_show: done (ok = %d).\n",ok);
	}

	if (out_file) (void) fclose (out_file);

	set_cursor (wid_main_window, DEFAULT_CURSOR);
    }
    return (done);
}

/**************************************************
**      wp_read_show
**************************************************/
static Boolean wp_read_show (XtPointer client_data)
{
    Boolean		    done, ok;
    char		    ch0[256], *cp1, *cp2, *cp3, *cp4;
    FILE		    *in_file;
    int			    len1;
    palette_entry_type	    *pep1;
    palette_type	    *color_palette;
    slide_body_type	    *bp1;
    String		    str1, str2;
    struct read_show_info   *dp;
    reclist_type	    *show, *master_list;
    unsigned int	    indx;

    done = False;
    ok = True;
    dp = (struct read_show_info *) client_data;
    color_palette = dp->color_palette_ptr;
    show = dp->show_ptr;
    master_list = dp->master_list_ptr;
    in_file = dp->filep;

    if (!in_file)
    {
	if (debug_flag)
	{
	    printf ("wp_read_show: initializing.\n");
	}

	set_cursor (wid_main_window, XC_watch);

	dp->new_slide = slide_create ();
	dp->new_element = element_create ();

	reclist_clear (show);
	reclist_clear (master_list);

	in_file = fopen (dp->filename, "r");
	dp->filep = in_file;
	ok = (in_file != (FILE *) NULL);
    }
    else
    {
	if (debug_flag)
	{
	    printf ("wp_read_show: reading slide %u.\n",
		    show->count);
	}

	if (fgets(ch0,sizeof(ch0),in_file))
	{
	    /* Remove newline from end of string. */
	    len1 = strlen (ch0);
	    if (len1)
	    {
	    	ch0 [len1 - 1] = '\0';
	    }
	    if (ch0[0] != '!')
	    {
		/*  Parse input line into fields, keeping a
		    separate copy of fields that may be
		    text. Otherwise, we would truncate the
		    text at the first ':' character. */
		cp1 = strchr (ch0, ':');
		if (cp1) *cp1++ = '\0'; else cp1 = "";
		str1 = XtNewString ((String) cp1);
		cp2 = strchr (cp1, ':');
		if (cp2) *cp2++ = '\0'; else cp2 = "";
		str2 = XtNewString ((String) cp2);
		cp3 = strchr (cp2, ':');
		if (cp3) *cp3++ = '\0'; else cp3 = "";
		cp4 = strchr (cp3, ':');
		if (cp4) *cp4++ = '\0'; else cp4 = "";

		if (strcmp (ch0, "BEGIN_SLIDE") == 0)
		{
		    slide_clear (dp->new_slide);
		}
		else if (strcmp (ch0, "END_SLIDE") == 0)
		{
		    reclist_insert_item (show,
					 0,
					 (void *) dp->new_slide);
		}
		else if (strcmp (ch0, "MASTER_NAME") == 0)
		{
		    XtFree (dp->new_slide->master_name);
		    dp->new_slide->master_name = XtNewString (str1);
		}
		else if (strcmp (ch0, "HEADER_TEXT") == 0)
		{
		    XtFree (dp->new_slide->header_text);
		    dp->new_slide->header_text = XtNewString (str1);
		}
		else if (strcmp (ch0, "HEADER_BAR") == 0)
		{
		    dp->new_slide->header_bar = (cp1[0] == 'Y');
		}
		else if (strcmp (ch0, "TITLE_TEXT") == 0)
		{
		    XtFree (dp->new_slide->title_text);
		    dp->new_slide->title_text = XtNewString (str1);
		}
		else if (strcmp (ch0, "BODY_TEXT") == 0)
		{
		    indx = atoi (cp1);
		    if ((indx) && (indx <= SLIDE_BODY_MAX))
		    {
			bp1 = &dp->new_slide->body_lines[indx - 1];
			XtFree (bp1->body_text);
			bp1->body_text = XtNewString (str2);
		    }
		}
		else if (strcmp (ch0, "BODY_BULLET") == 0)
		{
		    indx = atoi (cp1);
		    if ((indx) && (indx <= SLIDE_BODY_MAX))
		    {
			bp1 = &dp->new_slide->body_lines[indx - 1];
			bp1->body_bullet = (cp2[0] == 'Y');
		    }
		}

		else if (strcmp (ch0, "BEGIN_MASTER") == 0)
		{
		    dp->new_master =
			find_master (master_list,cp1,True);
		}
		else if (strcmp (ch0, "BEGIN_ELEMENT") == 0)
		{
		    element_clear (dp->new_element);
		    dp->new_element->element_type =
			lookup_el_type (cp1);
		}
		else if (strcmp (ch0, "ELEM_GEOMETRY") == 0)
		{
		    dp->new_element->element_x = atoi (cp1);
		    dp->new_element->element_y = atoi (cp2);
		    dp->new_element->element_width = atoi (cp3);
		    dp->new_element->element_height = atoi (cp4);
		}
		else if (strcmp (ch0, "ELEM_FONT") == 0)
		{
		    XtFree (dp->new_element->element_font_name);
		    dp->new_element->element_font_name =
			XtNewString (str1);
		}
		else if (strcmp (ch0, "ELEM_PIXMAP") == 0)
		{
		    XtFree (dp->new_element->element_pixmap_name);
		    dp->new_element->element_pixmap_name =
			XtNewString (str1);
		}
		else if (strcmp (ch0, "ELEM_FG") == 0)
		{
		    dp->new_element->element_foreground = atoi (cp1);
		}
		else if (strcmp (ch0, "ELEM_BG") == 0)
		{
		    dp->new_element->element_background = atoi (cp1);
		}
		else if (strcmp (ch0, "ELEM_DATA") == 0)
		{
		    dp->new_element->element_data = atoi (cp1);
		    dp->new_element->element_index = atoi (cp2);
		}
		else if (strcmp (ch0, "END_ELEMENT") == 0)
		{
		    if (dp->new_master)
		    {
			reclist_insert_item
			    (dp->new_master->elem_list,
			     0,
			     (void *) dp->new_element);
		    }
		}

		else if (strcmp (ch0, "COLOR") == 0)
		{
		    indx = atoi (cp1);
		    if ((indx) && (indx <= color_palette->size))
		    {
			pep1 = color_palette->entries + indx - 1;
			XtFree (pep1->color_name);
			pep1->color_name = XtNewString (str2);
			if (pep1->allocated_pixel)
			{
			    (void) store_color_in_cell
				(color_palette->screen,
				 pep1->color_name,
				 pep1->pixel);
			}
		    }
		}

		/*  Free the duplicate "text" strings we
		    allocated when we started looking at
		    this input line. */
		XtFree (str1);
		XtFree (str2);
	    }

	}
	else
	{
	    done = True;
	}
    }

    if (!ok) done = True;

    if (done)
    {
	if (debug_flag)
	{
	    printf ("wp_read_show: done (ok = %d).\n",ok);
	}

	if (in_file) (void) fclose (in_file);

	show->modified_flag = False;
	color_palette->modified_flag = False;
	master_list->modified_flag = False;

	if (dp->new_slide)
	{
	    slide_free (dp->new_slide);
	}

	XtFree ((char *) dp->new_element);

	if (master_list->count == 0)
	{
	    setup_default_master_list (master_list);
	}

	refresh_master_names (master_list);

	/*  Display the current color palette info. */
	update_paledit_info (color_palette);

	if (ok)
	{
	    set_file_name (dp->filename);
	}
	else
	{
	    display_error_text ("kcs_error_read_show");
	    set_file_name ("Untitled");
	}
	
	load_list_from_show (show);
	clear_entry ();
	update_show_status (show);
	set_cursor (wid_main_window, DEFAULT_CURSOR);
    }

    return (done);
}

/**************************************************
**      wp_write_show
**************************************************/
static Boolean wp_write_show (XtPointer client_data)
{
    Boolean		    done, ok;
    FILE		    *out_file;
    palette_type	    *color_palette;
    reclist_type	    *show, *master_list;
    struct write_show_info  *dp;

    dp = (struct write_show_info *) client_data;
    show = dp->show_ptr;
    color_palette = dp->color_palette_ptr;
    master_list = dp->master_list_ptr;
    out_file = dp->filep;
    done = False;
    ok = True;

    if (!out_file)
    {
	if (debug_flag)
	{
	    printf ("wp_write_show: initializing.\n");
	}

	set_cursor (wid_main_window, XC_watch);

	out_file = fopen (dp->filename, "w");
	dp->filep = out_file;
	ok = (out_file != (FILE *) NULL);
    }

    else if (dp->color_index < color_palette->size)
    {
	dp->color_index++;

	if (debug_flag)
	{
	    printf ("wp_write_show: writing palette entry %u.\n",
		    dp->color_index);
	}

	if (dp->color_index == 1)
	{
	    fprintf (out_file, "BEGIN_COLORS\n");
	}

	write_color (out_file, color_palette, dp->color_index);

	if (dp->color_index == color_palette->size)
	{
	    fprintf (out_file, "END_COLORS\n");
	}

    }

    else if (dp->master_index < master_list->count)
    {
	dp->master_index++;

	if (debug_flag)
	{
	    printf ("wp_write_show: writing master %u.\n",
		    dp->master_index);
	}

	if (dp->master_index == 1)
	{
	    fprintf (out_file, "BEGIN_MASTERS\n");
	}

	write_master (out_file, master_list, dp->master_index);

	if (dp->master_index == master_list->count)
	{
	    fprintf (out_file, "END_MASTERS\n");
	}

    }

    else if (dp->slide_index < show->count)
    {
	dp->slide_index++;

	if (debug_flag)
	{
	    printf ("wp_write_show: writing slide %u.\n",
		    dp->slide_index);
	}

	write_slide (out_file, show, dp->slide_index);
    }

    else
    {
	done = True;
    }

    if (!ok) done = True;

    if (done)
    {
	if (debug_flag)
	{
	    printf ("wp_write_show: done (ok = %d).\n",ok);
	}

	if (out_file) (void) fclose (out_file);

	if (ok)
	{
	    show->modified_flag = False;
	    color_palette->modified_flag = False;
	    master_list->modified_flag = False;
	    set_file_name (dp->filename);
	}
	set_cursor (wid_main_window, DEFAULT_CURSOR);
    }
    return (done);
}

/**************************************************
**      write_color
**************************************************/
static void write_color (FILE *out_file,
			 palette_type *pal_ptr,
			 unsigned int sindx)
{
    String	str1;

    str1 = pal_ptr->entries [sindx -1].color_name;

    if (str1)
	fprintf (out_file,
		 "COLOR:%u:%s\n",
		 sindx,
		 str1);
}
/**************************************************
**      write_master
**************************************************/
static void write_master (FILE *out_file,
			  reclist_type *mlp,
			  unsigned int mindx)
{
    master_element_type	    *ep;
    slide_master_type	    *mp;
    unsigned int	    eindx;

    mp = (slide_master_type *)
	reclist_get_item_ptr (mlp, mindx);
    fprintf (out_file, "BEGIN_MASTER:%s\n",
	     mp->name ? mp->name : "");

    for (eindx = 1;
	 eindx <= mp->elem_list->count;
	 eindx++)
    {
	ep = (master_element_type *)
	    reclist_get_item_ptr (mp->elem_list,eindx);
	fprintf (out_file,
		 "BEGIN_ELEMENT:%s\n",
		 element_type_names [ep->element_type]);
	fprintf (out_file,
		 "ELEM_GEOMETRY:%d:%d:%u:%u\n",
		 ep->element_x,
		 ep->element_y,
		 ep->element_width,
		 ep->element_height);
	if (ep->element_font_name)
	    fprintf (out_file,
		     "ELEM_FONT:%s\n",
		     ep->element_font_name);
	if (ep->element_pixmap_name)
	    fprintf (out_file,
		     "ELEM_PIXMAP:%s\n",
		     ep->element_pixmap_name);
	if (ep->element_foreground)
	    fprintf (out_file,
		     "ELEM_FG:%u\n",
		     ep->element_foreground);
	if (ep->element_background)
	    fprintf (out_file,
		     "ELEM_BG:%u\n",
		     ep->element_background);
	fprintf (out_file,
		 "ELEM_DATA:%d:%d\n",
		 ep->element_data,
		 ep->element_index);
	fprintf (out_file, "END_ELEMENT\n");
    }
    fprintf (out_file, "END_MASTER\n");
}

/**************************************************
**      write_show
**************************************************/
static void write_show (String filnam,
			reclist_type *show,
			palette_type *palette,
			reclist_type *master_list)
{
    struct write_show_info  *wpd_ptr;

    /*	Allocate and initialize data structure for
	work procedure. */
    wpd_ptr = XtNew (struct write_show_info);

    wpd_ptr->filename = XtNewString (filnam);
    wpd_ptr->show_ptr = show;
    wpd_ptr->color_palette_ptr = palette;
    wpd_ptr->master_list_ptr = master_list;
    wpd_ptr->filep = (FILE *) NULL;
    wpd_ptr->master_index = 0;
    wpd_ptr->color_index = 0;
    wpd_ptr->slide_index = 0;

    /*	Request work procedure. */
    (void) XtAppAddWorkProc
		(app_context,
		 (XtWorkProc) wp_write_show,
		 (XtPointer) wpd_ptr);
}

/**************************************************
**      write_slide
**************************************************/
static void write_slide (FILE *out_file,
			 reclist_type *show,
			 unsigned int sindx)
{
    slide_body_type	*bp;
    slide_info_type	*sp;
    unsigned int	bindx;

    sp = (slide_info_type *)
	reclist_get_item_ptr (show, sindx);
    fprintf (out_file, "BEGIN_SLIDE\n");

    if (sp->master_name)
	fprintf (out_file, "MASTER_NAME:%s\n",
			   sp->master_name);
    if (sp->header_text)
	fprintf (out_file, "HEADER_TEXT:%s\n",
			   sp->header_text);
    if (sp->header_bar)
	fprintf (out_file, "HEADER_BAR:Y\n");
    if (sp->title_text)
	fprintf (out_file, "TITLE_TEXT:%s\n",sp->title_text);

    bp = sp->body_lines;
    for (bindx = 1;  bindx <= SLIDE_BODY_MAX;  bp++,bindx++)
    {
	if (bp->body_text)
	    fprintf (out_file,
		     "BODY_TEXT:%u:%s\n",
		     bindx,
		     bp->body_text);
	if (bp->body_bullet)
	    fprintf (out_file,"BODY_BULLET:%u:Y\n",bindx);
    }

    fprintf (out_file, "END_SLIDE\n");
}
